import json
from logging import DEBUG, ERROR, INFO
from typing import Dict, Optional
import click
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.api.metadata import update_snapshot_and_timestamp, update_target_metadata
from taf.models.types import RolesIterator
from taf.models.types import RolesKeysData
from taf.api.utils import check_if_clean, commit_and_push
from taf.models.converter import from_dict

import taf.repositoriesdb as repositoriesdb
from pathlib import Path
from taf.api.roles import (
    create_delegations,
    _initialize_roles_and_keystore,
    setup_role,
)
from taf.api.targets import register_target_files

from taf.auth_repo import AuthenticationRepository
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.git import GitRepository
from taf.keys import load_sorted_keys_of_new_roles
from taf.repository_tool import Repository
from tuf.repository_tool import create_new_repository
from taf.log import taf_logger


@log_on_start(
    DEBUG, "Adding or updating dependency {dependency_name:s}", logger=taf_logger
)
@log_on_end(DEBUG, "Finished adding or updating dependency", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding a new dependency: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_dependency(
    path: str,
    dependency_name: str,
    branch_name: str,
    out_of_band_commit: str,
    keystore: str,
    dependency_path: Optional[str] = None,
    library_dir: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    custom: Optional[Dict] = None,
    prompt_for_keys: Optional[bool] = False,
    commit: Optional[bool] = True,
) -> None:
    """
    Add a dependency (an authentication repository) to dependencies.json or update it if it was already added to this file.
    Information that is added to dependencies.json includes out-of-band authentication commit name of the branch which contains
    that commit. It is possible for multiple branches to contain the same commit, so it's important to list branch name as well.
    If repository exists on disk, this function also validates that out_of_brand_commit and branch_name exist and that the
    commit belongs to the specified branch.

    Arguments:
        path: Path to the authentication repository.
        dependency_name: Name of the dependency.
        branch_name: Name of the branch which contains the out-of-band authentication commit.
        out_of_band_commit: SHA of out-of-band authentication commit.
        keystore: Location of the keystore files.
        dependency_path (optional): Path to the dependency repository which is to be added. Can be omitted if dependency_name
        library_dir (optional): Path to the library's root directory. Determined based on the authentication repository's path if not provided.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        custom (optional): Additional data that will be added to dependencies.json if specified.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.

    Side Effects:
        Updates dependencies.json, targets, snapshot and timestamp metadata, writes changes to disk
        and commits them.

    Raises:
        TAFError if dependency exists on disk, but does not contain specified commit and/or branch

    Returns:
        None
    """
    if path is None:
        raise TAFError("Authentication repository's path not provided")

    auth_repo = AuthenticationRepository(path=path)
    if not auth_repo.is_git_repository_root:
        print(f"{path} is not a git repository!")
        return
    if library_dir is None:
        library_dir = auth_repo.path.parent.parent

    if dependency_path is not None:
        dependency = GitRepository(path=dependency_path)
    else:
        dependency = GitRepository(library_dir, dependency_name)

    if dependency.is_git_repository:
        branch_name, out_of_band_commit = _determine_out_of_band_data(
            dependency, branch_name, out_of_band_commit
        )
    else:
        if branch_name is None or out_of_band_commit is None:
            raise TAFError(
                "Branch name and out-of-band commit must be specified if repository does not exist on disk"
            )
        if not click.confirm(
            "Dependency not on disk. Proceed without validating branch and commit?"
        ):
            return

    # add to dependencies.json or update the entry
    dependencies_json = repositoriesdb.load_dependencies_json(auth_repo)

    # if dependencies.json does not exist, initialize it
    if not dependencies_json:
        dependencies_json = {"dependencies": {}}

    dependencies = dependencies_json["dependencies"]
    if dependency_name in dependencies:
        print(f"{dependency_name} already added to dependencies.json. Overwriting")
    dependencies[dependency_name] = {
        "out-of-band-authentication": out_of_band_commit,
        "branch": branch_name,
    }
    if custom:
        dependencies[dependency_name]["custom"] = custom

    # update content of repositories.json before updating targets metadata
    Path(auth_repo.path, repositoriesdb.DEPENDENCIES_JSON_PATH).write_text(
        json.dumps(dependencies_json, indent=4)
    )

    removed_targets_data = {}
    added_targets_data = {repositoriesdb.DEPENDENCIES_JSON_NAME: {}}
    update_target_metadata(
        auth_repo,
        added_targets_data,
        removed_targets_data,
        keystore,
        write=False,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
    )

    # update snapshot and timestamp calls write_all, so targets updates will be saved too
    update_snapshot_and_timestamp(
        auth_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    )
    if commit:
        commit_and_push(auth_repo)
    else:
        print("\nPlease commit manually.\n")


@log_on_start(
    INFO, "Creating a new authentication repository {path:s}", logger=taf_logger
)
@log_on_end(INFO, "Finished creating a new repository", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while creating a new repository: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def create_repository(
    path: str,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
    commit: Optional[bool] = False,
    test: Optional[bool] = False,
) -> None:
    """
    Create a new authentication repository. Generate initial metadata files.
    If target files already exist, add corresponding targets information to
    targets metadata files.

    Arguments:
        path: Authentication repository's location.
        keystore: Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        test (optional): Specifies if the created repository is a test authentication repository.

    Side Effects:
        Creates a new authentication repository (initializes a new git repository and generates tuf metadata)

    Returns:
        None
    """
    auth_repo = AuthenticationRepository(path=path)
    path = Path(path)

    if not _check_if_can_create_repository(auth_repo):
        return

    roles_key_infos, keystore = _initialize_roles_and_keystore(
        roles_key_infos, keystore
    )

    roles_keys_data = from_dict(roles_key_infos, RolesKeysData)
    repository = create_new_repository(str(auth_repo.path))
    signing_keys, verification_keys = load_sorted_keys_of_new_roles(
        auth_repo=auth_repo,
        roles=roles_keys_data.roles,
        yubikeys_data=roles_keys_data.yubikeys,
        keystore=keystore,
    )
    # set threshold and register keys of main roles
    # we cannot do the same for the delegated roles until delegations are created
    for role in RolesIterator(roles_keys_data.roles, include_delegations=False):
        setup_role(
            role,
            repository,
            verification_keys[role.name],
            signing_keys.get(role.name),
        )

    create_delegations(
        roles_keys_data.roles.targets, repository, verification_keys, signing_keys
    )

    # if the repository is a test repository, add a target file called test-auth-repo
    if test:
        test_auth_file = (
            Path(auth_repo.path, auth_repo.targets_path) / auth_repo.TEST_REPO_FLAG_FILE
        )
        test_auth_file.touch()

    # register and sign target files (if any)
    taf_repository = Repository(path)
    taf_repository._tuf_repository = repository
    updated = register_target_files(
        path,
        keystore,
        roles_key_infos,
        commit=False,
        taf_repo=taf_repository,
        write=True,
    )
    if not updated:
        repository.writeall()

    if commit:
        auth_repo.init_repo()
        commit_and_push(auth_repo, push=False)
    else:
        print("\nPlease commit manually.\n")


def _check_if_can_create_repository(auth_repo: AuthenticationRepository) -> bool:
    """
    Check if a new authentication repository can be created at the specified location.
    A repository can be created if there is not directory at the repository's location
    or if it does exists, is not the root of a git repository.

    Arguments:
        auth_repo: Authentication repository.

    Side Effects:
        None

    Returns:
        True if a new authentication repository can be created, False otherwise.
    """
    path = Path(auth_repo.path)
    if path.is_dir():
        # check if there is non-empty metadata directory
        if auth_repo.metadata_path.is_dir() and any(auth_repo.metadata_path.iterdir()):
            if auth_repo.is_git_repository:
                print(
                    f'"{path}" is a git repository containing the metadata directory. Generating new metadata files could make the repository invalid. Aborting.'
                )
                return False
            if not click.confirm(
                f'Metadata directory found inside "{path}". Recreate metadata files?'
            ):
                return False
    return True


def _determine_out_of_band_data(
    dependency: GitRepository, branch_name: str, out_of_band_commit: str
):
    """
    Determines values of out-of-band branch and commit as a part of adding a new
    dependency to dependencies.json. If not defined, branch is set to the default branch
    of the repository and commit to the first commit of that branch.
    """
    is_branch_specified = branch_name is not None
    is_commit_specified = out_of_band_commit is not None

    if branch_name is None:
        branch_name = dependency.default_branch
    else:
        if not dependency.branch_exists(branch_name):
            raise TAFError(f"Branch {branch_name} does not exists in {dependency.name}")

    if out_of_band_commit is None:
        out_of_band_commit = dependency.get_first_commit_on_branch(branch_name)
    else:
        try:
            branches = dependency.branches_containing_commit(
                out_of_band_commit, strip_remote=True
            )
        except TAFError:
            raise TAFError("Specified out-of-band authentication commit does not exist")
        if branch_name not in branches:
            raise TAFError(
                f"Commit {out_of_band_commit} not on branch {dependency.branch_name}"
            )

    if not is_branch_specified or not is_commit_specified:
        if not click.confirm(
            f"Branch and out-of-band authentication commit will be set to {branch_name} and {out_of_band_commit}. Proceed?"
        ):
            return

    return branch_name, out_of_band_commit


@log_on_start(DEBUG, "Remove dependency {dependency_name:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished removing dependency", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while removing a dependency: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def remove_dependency(
    path: str,
    dependency_name: str,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    commit: Optional[bool] = True,
) -> None:
    """
    Remove a dependency (an authentication repository) from dependencies.json

    Arguments:
        path: Path to the authentication repository.
        dependency_name: Name of the dependency which should be removed.
        keystore: Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.

    Side Effects:
        Updates dependencies.json, targets, snapshot and timestamp metadata, writes changes to disk
        and commits them.

    Returns:
        None
    """
    if path is None:
        raise TAFError("Authentication repository's path not provided")

    auth_repo = AuthenticationRepository(path=path)
    if not auth_repo.is_git_repository_root:
        print(f"{path} is not a git repository!")
        return

    # add to dependencies.json or update the entry
    dependencies_json = repositoriesdb.load_dependencies_json(auth_repo)

    if not dependencies_json:
        print("dependencies.json does not exist")
        return

    dependencies = dependencies_json["dependencies"]

    if dependency_name not in dependencies:
        print("Dependency not in dependencies.json")
        return

    dependencies.pop(dependency_name)

    # update content of repositories.json before updating targets metadata
    Path(auth_repo.path, repositoriesdb.DEPENDENCIES_JSON_PATH).write_text(
        json.dumps(dependencies_json, indent=4)
    )

    removed_targets_data = {}
    added_targets_data = {repositoriesdb.DEPENDENCIES_JSON_NAME: {}}
    update_target_metadata(
        auth_repo,
        added_targets_data,
        removed_targets_data,
        keystore,
        write=False,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
    )

    # update snapshot and timestamp calls write_all, so targets updates will be saved too
    update_snapshot_and_timestamp(
        auth_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    )
    if commit:
        commit_and_push(auth_repo)
    else:
        print("\nPlease commit manually.\n")



import datetime
from tuf.ngclient._internal import trusted_metadata_set


class GitTrustedMetadataSet(trusted_metadata_set.TrustedMetadataSet):
    """
    This class represents a "divergence" from TUF metadata validation.
    TUF, by design, checks if metadata expired by validating "expiration_date" field during "refresh".
    As of TUF 2.0.0, this means that "expiration_date" from the loaded metadata is compared to current time,
    which only makes sense when we're validating the latest metadata. However,
    TAF validates metadata across history. We do not want to validate expiration for each commit (revision),
    since it will always be considered "expired".
    Instead, for each revision in commit history we override the "reference_time" attribute so that
    past metadata will not be considered expired.

    See: GitUpdater
    """

    def __init__(self, data):
        super(GitTrustedMetadataSet, self).__init__(data)
        self.reference_time = datetime.datetime.min



from datetime import datetime
from logging import DEBUG, ERROR, INFO
from pathlib import Path
from typing import Dict, List, Optional
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.api.utils import check_if_clean, commit_and_push
from taf.exceptions import TAFError
from taf.git import GitRepository
from taf.keys import load_signing_keys
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.repository_tool import Repository, is_delegated_role
from taf.log import taf_logger


@log_on_error(
    ERROR,
    "An error occurred while checking expiration dates: {e!r}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=False,
)
def check_expiration_dates(
    path: str,
    interval: Optional[int] = None,
    start_date: Optional[datetime] = None,
    excluded_roles: Optional[List[str]] = None,
) -> None:
    """
    Check if any metadata files (roles) are expired or will expire in the next <interval> days.
    Prints a list of expired roles.

    Arguments:
        path: Authentication repository's location.
        interval: Number of days ahead to check for expiration.
        start_date: Date from which to start checking for expiration.
        excluded_roles: List of roles to exclude from the check.

    Side Effects:
        Prints lists of roles that expired or are about to expire.

    Returns:
        None
    """
    path = Path(path)
    taf_repo = Repository(path)

    expired_dict, will_expire_dict = taf_repo.check_roles_expiration_dates(
        interval, start_date, excluded_roles
    )

    if expired_dict or will_expire_dict:
        now = datetime.now()
        print(
            f"Given a {interval} day interval from today ({start_date.strftime('%Y-%m-%d')}):"
        )
        for role, expiry_date in expired_dict.items():
            delta = now - expiry_date
            print(
                f"{role} expired {delta.days} days ago (on {expiry_date.strftime('%Y-%m-%d')})"
            )
        for role, expiry_date in will_expire_dict.items():
            delta = expiry_date - now
            print(
                f"{role} will expire in {delta.days} days (on {expiry_date.strftime('%Y-%m-%d')})"
            )
    else:
        print(f"No roles will expire within the given {interval} day interval")


@check_if_clean
def update_metadata_expiration_date(
    path: str,
    roles: List[str],
    interval: int,
    keystore: Optional[str] = None,
    scheme: Optional[str] = None,
    start_date: Optional[datetime] = None,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Update expiration dates of the specified roles and all other roles that need
    to be signed in order to guarantee validity of the repository e.g. snapshot
    and timestamp need to be signed after a targets role is updated.

    Arguments:
        path: Authentication repository's location.
        roles: A list of roles whose expiration dates should be updated.
        interval: Number of days added to the start date in order to calculate the
            expiration date.
        keystore (optional): Keystore directory's path
        scheme (optional): Signature scheme.
        start_date (optional): Date to which expiration interval is added.
            Set to today if not specified.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.

    Side Effects:
        Updates metadata files, saves changes to disk and commits changes
        unless no_commit is set to True.

    Returns:
        None
    """
    if start_date is None:
        start_date = datetime.now()

    taf_repo = Repository(path)
    loaded_yubikeys = {}
    roles_to_update = []

    if "root" in roles:
        roles_to_update.append("root")
    if "targets" in roles:
        roles_to_update.append("targets")
    for role in roles:
        if is_delegated_role(role):
            roles_to_update.append(role)

    if len(roles_to_update) or "snapshot" in roles:
        roles_to_update.append("snapshot")
    roles_to_update.append("timestamp")

    for role in roles_to_update:
        _update_expiration_date_of_role(
            taf_repo,
            role,
            loaded_yubikeys,
            keystore,
            start_date,
            interval,
            scheme,
            prompt_for_keys,
        )

    if not commit:
        print("\nPlease commit manually. \n")
    else:
        auth_repo = GitRepository(path=path)
        commit_and_push(auth_repo)


@log_on_end(INFO, "Updated expiration date of {role:s}", logger=taf_logger)
@log_on_error(
    ERROR,
    "Error: could not update expiration date: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def _update_expiration_date_of_role(
    auth_repo: Repository,
    role: str,
    loaded_yubikeys: Dict,
    keystore: str,
    start_date: datetime,
    interval: int,
    scheme: str,
    prompt_for_keys: bool,
) -> None:
    keys, yubikeys = load_signing_keys(
        auth_repo,
        role,
        loaded_yubikeys=loaded_yubikeys,
        keystore=keystore,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
    )
    # sign with keystore
    if len(keys):
        auth_repo.update_role_keystores(
            role, keys, start_date=start_date, interval=interval
        )
    if len(yubikeys):  # sign with yubikey
        auth_repo.update_role_yubikeys(
            role, yubikeys, start_date=start_date, interval=interval
        )


@log_on_end(INFO, "Updated snapshot and timestamp", logger=taf_logger)
@log_on_error(
    ERROR,
    "Could not update snapshot and timestamp: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def update_snapshot_and_timestamp(
    taf_repo: Repository,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    write_all: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Sign snapshot and timestamp metadata files.

    Arguments:
        taf_repo: Authentication repository.
        keystore: Keystore directory's path.
        scheme (optional): Signature scheme.
        write_all (optional): If True, writes authentication repository's
            changes to disk.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.

    Side Effects:
        Updates metadata files, saves changes to disk if write_all is True

    Returns:
        None
    """
    loaded_yubikeys = {}

    for role in ("snapshot", "timestamp"):
        keystore_keys, yubikeys = load_signing_keys(
            taf_repo,
            role,
            keystore,
            loaded_yubikeys,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
        )
        if len(yubikeys):
            update_method = taf_repo.roles_yubikeys_update_method(role)
            update_method(yubikeys, write=False)
        if len(keystore_keys):
            update_method = taf_repo.roles_keystore_update_method(role)
            update_method(keystore_keys, write=False)

    if write_all:
        taf_repo.writeall()


@log_on_start(DEBUG, "Updating target metadata", logger=taf_logger)
@log_on_end(DEBUG, "Updated target metadata", logger=taf_logger)
@log_on_error(
    ERROR,
    "Could not update target metadata: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def update_target_metadata(
    taf_repo: Repository,
    added_targets_data: Dict,
    removed_targets_data: Dict,
    keystore: str,
    write: Optional[bool] = False,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """Given dictionaries containing targets that should be added and targets that should
    be removed, update and sign target metadata files and, if write is True, also
    sign snapshot and timestamp.

    Sign snapshot and timestamp metadata files

    Arguments:
        taf_repo: Authentication repository.
        added_targets_data(dict): Dictionary containing targets data that should be added.
        removed_targets_data(dict): Dictionary containing targets data that should be removed.
        keystore: Keystore directory's path.
        write (optional): If True, updates snapshot and timestamp and write changes to disk.
        scheme (optional): Signature scheme.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.

    Side Effects:
        Updates metadata files, saves changes to disk if write_all is True

    Returns:
        True if there were targets that were updated, False otherwise
    """

    added_targets_data = {} if added_targets_data is None else added_targets_data
    removed_targets_data = {} if removed_targets_data is None else removed_targets_data

    roles_targets = taf_repo.roles_targets_for_filenames(
        list(added_targets_data.keys()) + list(removed_targets_data.keys())
    )

    if not roles_targets:
        taf_logger.info("No target files to sign")
        return False

    # update targets
    loaded_yubikeys = {}
    for role, target_paths in roles_targets.items():
        keystore_keys, yubikeys = load_signing_keys(
            taf_repo,
            role,
            keystore,
            loaded_yubikeys,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
        )
        targets_data = dict(
            added_targets_data={
                path: val
                for path, val in added_targets_data.items()
                if path in target_paths
            },
            removed_targets_data={
                path: val
                for path, val in removed_targets_data.items()
                if path in target_paths
            },
        )

        if len(yubikeys):
            taf_repo.update_targets_yubikeys(yubikeys, write=False, **targets_data)
        if len(keystore_keys):
            taf_repo.update_targets_keystores(
                keystore_keys, write=False, **targets_data
            )

    if write:
        update_snapshot_and_timestamp(
            taf_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
        )
    return True



from attrs import define, field
from typing import Dict


@define
class Update:
    changed: bool = field(default=False)
    event: str = field(default="")
    error_msg: str = field(default="")
    auth_repos: Dict = field(factory=dict)
    auth_repo_name: str = field(default="")



import click
import errno
import datetime
import time
import json
import os
import stat
import subprocess
import tempfile
import shutil
import uuid
from getpass import getpass
from functools import wraps
from pathlib import Path
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import (
    load_pem_public_key,
    load_pem_private_key,
)
from json import JSONDecoder
import taf.settings
from taf.exceptions import PINMissmatchError
from taf.log import taf_logger


def _iso_parse(date):
    return datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S.%f")


class IsoDateParamType(click.ParamType):
    name = "iso_date"

    def convert(self, value, param, ctx):
        if value is None:
            return datetime.datetime.now()

        if isinstance(value, datetime.datetime):
            return value
        try:
            return _iso_parse(value)
        except ValueError as ex:
            self.fail(str(ex), param, ctx)


ISO_DATE_PARAM_TYPE = IsoDateParamType()


def extract_x509(cert_pem):
    cert = x509.load_pem_x509_certificate(cert_pem, default_backend())

    def _get_attr(oid):
        attrs = cert.subject.get_attributes_for_oid(oid)
        return attrs[0].value if len(attrs) > 0 else ""

    return {
        "name": _get_attr(x509.OID_COMMON_NAME),
        "organization": _get_attr(x509.OID_ORGANIZATION_NAME),
        "country": _get_attr(x509.OID_COUNTRY_NAME),
        "state": _get_attr(x509.OID_STATE_OR_PROVINCE_NAME),
        "locality": _get_attr(x509.OID_LOCALITY_NAME),
        "valid_from": cert.not_valid_before.strftime("%Y-%m-%d"),
        "valid_to": cert.not_valid_after.strftime("%Y-%m-%d"),
    }


def get_cert_names_from_keyids(certs_dir, keyids):
    cert_names = []
    for keyid in keyids:
        try:
            name = extract_x509((Path(certs_dir) / keyid + ".pem").read_bytes())["name"]
            if not name:
                print("Cannot extract common name from x509, using key id instead.")
                cert_names.append(keyid)
            else:
                cert_names.append(name)
        except FileNotFoundError:
            print(f"Certificate does not exist ({keyid}).")
    return cert_names


def get_key_size(key_pem_path, decrypt_pwd=None):
    try:
        key = load_pem_public_key(Path(key_pem_path).read_bytes(), default_backend())
    except ValueError:
        key = load_pem_private_key(
            Path(key_pem_path).read_bytes(), decrypt_pwd, default_backend()
        )
    except Exception:
        return 0

    return key.key_size


def get_pin_for(name, confirm=True, repeat=True):
    pin = getpass(f"Enter PIN for {name}: ")
    if confirm:
        if pin != getpass(f"Confirm PIN for {name}: "):
            err_msg = "PINs don't match!"
            if repeat:
                print(err_msg)
                get_pin_for(name, confirm, repeat)
            else:
                raise PINMissmatchError(err_msg)
    return pin


def extract_json_objects_from_trusted_stdout(text, decoder=JSONDecoder()):
    """Find JSON objects in text, and yield the decoded JSON data

    Does not attempt to look for JSON arrays, text, or other JSON types outside
    of a parent JSON object.

    """
    pos = 0
    while True:
        match = text.find("{", pos)
        if match == -1:
            break
        try:
            result, index = decoder.raw_decode(text[match:])
            yield result
            pos = match + index
        except ValueError:
            pos = match + 1


def read_input_dict(value):
    if value is None:
        return {}
    if type(value) is str:
        if Path(value).is_file():
            with open(value) as f:
                try:
                    value = json.loads(f.read())
                except json.decoder.JSONDecodeError:
                    print(f"\nWARNING: {value} is not a valid json!\n")
                    return {}

        else:
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError:
                print(f"\nWARNING: {value} is not a valid json!\n")
                return {}

    return value


def run(*command, **kwargs):
    """Run a command and return its output. Call with `debug=True` to print to
    stdout.
    In order to get bytes, call this command with `raw=True` argument.
    """
    # Skip decoding
    raw = kwargs.pop("raw", False)
    data = kwargs.pop("input", None)

    if len(command) == 1 and isinstance(command[0], str):
        command = command[0].split()
    if taf.settings.LOG_COMMAND_OUTPUT:
        taf_logger.debug("About to run command {}", " ".join(command))

    def _format_word(word, **env):
        """To support words such as @{u} needed for git commands."""
        try:
            return word.format(env)
        except KeyError:
            return word

    command = [_format_word(word, **os.environ) for word in command]
    try:
        options = dict(stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True)
        if not raw:
            options.update(universal_newlines=True)
        if data is not None:
            options.update(input=data)

        options.update(kwargs)
        completed = subprocess.run(command, **options)
    except subprocess.CalledProcessError as err:
        if err.stdout:
            taf_logger.debug(err.stdout)
        if err.stderr:
            taf_logger.debug(err.stderr)
        taf_logger.debug(
            "Command {} returned non-zero exit status {} with output {}",
            " ".join(command),
            err.returncode,
            err.output,
        )
        raise err

    if completed.stdout:
        if taf.settings.LOG_COMMAND_OUTPUT:
            taf_logger.debug(completed.stdout)

    if completed.returncode != 0:
        return None

    return completed.stdout if raw else completed.stdout.rstrip()


def normalize_file_line_endings(file_path):
    with open(file_path, "rb") as open_file:
        content = open_file.read()
    replaced_content = normalize_line_endings(content)
    if replaced_content != content:
        with open(file_path, "wb") as open_file:
            open_file.write(replaced_content)


def normalize_line_endings(file_content):
    WINDOWS_LINE_ENDING = b"\r\n"
    UNIX_LINE_ENDING = b"\n"
    replaced_content = file_content.replace(
        WINDOWS_LINE_ENDING, UNIX_LINE_ENDING
    ).rstrip(UNIX_LINE_ENDING)
    return replaced_content


def on_rm_error(_func, path, _exc_info):
    """Used by when calling rmtree to ensure that readonly files and folders
    are deleted.
    """
    try:
        os.chmod(path, stat.S_IWRITE)
    except OSError as e:
        taf_logger.debug(f"File at path {path} not found, error trace - {e}")
        return
    try:
        os.unlink(path)
    except (OSError, PermissionError) as e:
        taf_logger.debug(
            "WARNING: Failed to clean up temporary update files: {}. This is a known issue when running TAF in a subprocess. You could consider upgrading taf to see if cleanup errors persist",
            e,
        )
        pass


def safely_save_json_to_disk(data, permanent_path):
    tfile = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
    if data is not None:
        json.dump(data, tfile, indent=4)
    else:
        Path(permanent_path).write_text("")

    temp_file_path = tfile.name
    tfile.close()
    safely_move_file(temp_file_path, permanent_path, overwrite=True)


def safely_move_file(src, dst, overwrite=False):
    """Rename a file from ``src`` to ``dst``.

    *   Moves must be atomic.  ``shutil.move()`` is not atomic.
        Note that multiple threads may try to write to the cache at once,
        so atomicity is required to ensure the serving on one thread doesn't
        pick up a partially saved image from another thread.

    *   Moves must work across filesystems.  Often temp directories and the
        cache directories live on different filesystems.  ``os.rename()`` can
        throw errors if run across filesystems.

    So we try ``os.rename()``, but if we detect a cross-filesystem copy, we
    switch to ``shutil.move()`` with some wrappers to make it atomic.
    https://alexwlchan.net/2019/03/atomic-cross-filesystem-moves-in-python/
    MIT licensed.
    """
    src = str(src)
    dst = str(dst)
    if overwrite and os.path.isfile(dst):
        os.remove(dst)
    try:
        os.rename(src, dst)
    except OSError as err:
        if err.errno == errno.EXDEV:
            # Generate a unique ID, and copy `<src>` to the target directory
            # with a temporary name `<dst>.<ID>.tmp`.  Because we're copying
            # across a filesystem boundary, this initial copy may not be
            # atomic.  We intersperse a random UUID so if different processes
            # are copying into `<dst>`, they don't overlap in their tmp copies.
            copy_id = uuid.uuid4()
            tmp_dst = f"{dst}.{copy_id}.tmp"
            shutil.copyfile(src, tmp_dst)
            # Then do an atomic rename onto the new name, and clean up the
            # source image.
            os.rename(tmp_dst, dst)
            os.unlink(src)
        else:
            raise


def to_tuf_datetime_format(start_date, interval):
    """Used to convert datetime to format used while writing metadata:
    e.g. "2020-05-29T21:59:34Z",
    """
    datetime_object = start_date + datetime.timedelta(interval)
    datetime_object = datetime_object.replace(microsecond=0)
    return datetime_object.isoformat() + "Z"


class timed_run:
    """Decorator to let us capture the elapsed time and optionally print a timer and start/end
    messages around function calls"""

    def __init__(self, start_message=None, end_message="  completed in {} seconds"):
        self.start_message = start_message
        self.end_message = end_message
        self.start_time = None
        self.elapsed_time = None

    def start(self):
        self.start_time = time.time()
        if self.start_message is not None:
            print(self.start_message)

    def end(self):
        self.elapsed_time = time.time() - self.start_time
        if self.end_message is not None:
            print(self.end_message.format(int(self.elapsed_time)))

    def __call__(self, orig_func=None):
        @wraps(orig_func)
        def wrapper_func(*args, **kwargs):
            self.start()
            result = orig_func(*args, **kwargs) if orig_func else None
            self.end()
            return result

        return wrapper_func



from ast import Dict
from typing import Type
import attrs
import cattrs
from cattrs.gen import make_dict_structure_fn
from taf.exceptions import RolesKeyDataConversionError

converter = cattrs.Converter()
converter.register_structure_hook_factory(
    attrs.has,
    lambda cl: make_dict_structure_fn(
        cl, converter, _cattrs_forbid_extra_keys=True, _cattrs_detailed_validation=False
    ),
)


def from_dict(data_dict: Dict, model_type: Type):
    try:
        return converter.structure(data_dict, model_type)
    except cattrs.errors.IterableValidationError as e:
        raise RolesKeyDataConversionError(e.exceptions)
    except ValueError as e:
        raise RolesKeyDataConversionError([str(e)])



import os
import shutil
import tempfile
from functools import wraps
from pathlib import Path

from tuf.ngclient._internal import trusted_metadata_set
from taf.exceptions import GitError
from taf.log import taf_logger
import taf.settings as settings
from taf.auth_repo import AuthenticationRepository
from taf.exceptions import UpdateFailedError
from taf.utils import on_rm_error
from taf.updater.git_trusted_metadata_set import GitTrustedMetadataSet

from tuf.ngclient.fetcher import FetcherInterface
from tuf.api.exceptions import DownloadHTTPError


class GitUpdater(FetcherInterface):
    """
    This class implements parts of the update process specific to keeping
    metadata files and targets in a git repository. The vast majority of the
    update process is handled by TUF's updater. This class does not modify
    and of TUF's validation functionalities - it only handles loading of
    files.
    One substantial difference between our system and how TUF is designed lies in the
    fact that we want to validate every commit and not just check if the most recent
    one contains valid metadata. We want to check if the metadata was valid at the time
    when it was committed.

    Since we do not want to change the clients repository until
    we know it is safe to use git pull, the update works as follows:
    - The repository containing the metadata files is cloned to the temp folder.
    - The most recent commit in the client's local repository is determined. Based
    on that all new commits (newer than the client's most recent one in the
    cloned repository) are found.
    - A commit is considered to be a TUF Updater instance. We keep track of the current commit.
    - This class is designed in such a way that for each subsequent call of the
    updater's refresh method the metadata from next commit is used within TUF's updater.
    - The updater's method for downloading and retrieving current metadata is overriden
    by our own '_fetch' call. We override TUF's FetcherInterface abstract class to fetch
    metadata from local git revisions, instead of by downloading data from another protocol,
    like http/https. So, what we want to do is to return the current commit,
    and just the current commit. This means that that an exception will be raised
    if the version of the metadata file at that commit is not valid.
    - The same logic is used to handle targets.


    Attributes:
        - repository_directory: the client's local repository's location
        - metadata_dir_path: path of the metadata directory needed by the updater.
        - validation_auth_repo: a fresh clone of the metadata repository. It is
        a bare git repository. An instance of the `BareGitRepo` class.
        - users_auth_repo: an instance of the `GitRepo` class. The user's current
        git repository.
        - commits: a list of commits, starting with the most recent commit in the
        user's repository. The following commits are those committed after the
        the top one if the client's repo.
        - commits_indexes: a dictionary which stores index of the current commit
        per metadata file. The reason for separating the metadata files is that
        not all files are updated at the same time.
    """

    @property
    def current_commit(self):
        return self.commits[self.current_commit_index]

    @property
    def previous_commit(self):
        return self.commits[self.current_commit_index - 1]

    @property
    def metadata_dir(self):
        return str(self.metadata_dir_path)

    @property
    def targets_dir(self):
        return str(self.validation_auth_repo.path / "targets")

    def __init__(self, auth_url, repository_directory, repository_name):
        """
        Args:
        auth_url: repository url of the git repository which we want to clone.
        repository_directory: the client's local repository's location
        repository_name: name of the repository in 'organization/namespace' format.
        """
        self._original_tuf_trusted_metadata_set = (
            trusted_metadata_set.TrustedMetadataSet
        )
        self._patch_tuf_metadata_set(GitTrustedMetadataSet)

        conf_directory_root = settings.conf_directory_root
        default_branch = settings.default_branch
        validation_path = settings.validation_repo_path

        self.set_validation_repo(validation_path, auth_url)

        # users_auth_repo is the authentication repository
        # located on the users machine which needs to be updated
        self.users_auth_repo = AuthenticationRepository(
            repository_directory,
            repository_name,
            default_branch=default_branch,
            urls=[auth_url],
            conf_directory_root=conf_directory_root,
        )

        repository_directory = Path(self.users_auth_repo.path)

        if repository_directory.exists():
            if not self.users_auth_repo.is_git_repository_root:
                if len(list(repository_directory.glob("*"))):
                    raise UpdateFailedError(
                        f"{repository_directory} is not a git repository and is not empty"
                    )

        self._init_commits()

        self.repository_directory = str(repository_directory)

        tmp_dir = tempfile.mkdtemp()
        metadata_path = Path(tmp_dir, "metadata")
        metadata_path.mkdir(parents=True, exist_ok=True)

        self.metadata_dir_path = metadata_path

        try:
            self._init_metadata()
        except Exception:
            self.cleanup()
            raise UpdateFailedError(
                "Could not load metadata. Check if the URL and filesystem paths are correct."
            )

    def _fetch(self, url):
        try:
            return [
                self.validation_auth_repo.get_file(self.current_commit, url, raw=True)
            ]
        except GitError:
            raise DownloadHTTPError(f"Could not find {url}", status_code=404)
        except Exception as e:
            taf_logger.error(
                f"Retrieval of data at current revision {self.current_commit} failed with error - {str(e)}"
            )
            raise e

    def _init_commits(self):
        """
        Given a client's local repository which needs to be updated, creates a list
        of commits of the authentication repository newer than the most recent
        commit of the client's repository. These commits need to be validated.
        If the client's repository does not exist, all commits should be validated.
        We have to presume that the initial metadata is correct though (or at least
        the initial root.json).
        """
        # TODO check if users authentication repository is clean
        # load the last validated commit from the conf file
        if settings.overwrite_last_validated_commit:
            last_validated_commit = settings.last_validated_commit
        else:
            last_validated_commit = self.users_auth_repo.last_validated_commit

        try:
            commits_since = self.validation_auth_repo.all_commits_since_commit(
                last_validated_commit
            )
        except GitError as e:
            if "Invalid revision range" in str(e):
                taf_logger.error(
                    "Commit {} is not contained by the remote repository {}.",
                    last_validated_commit,
                    self.validation_auth_repo.name,
                )
                raise UpdateFailedError(
                    f"Commit {last_validated_commit} is no longer contained by repository"
                    f" {self.validation_auth_repo.name}. This could "
                    "either mean that there was an unauthorized push to the remote "
                    "repository, or that last_validated_commit file was modified."
                )
            else:
                raise e

        # Check if the user's head commit matches the saved one
        # That should always be the case
        # If it is not, it means that someone, accidentally or maliciously made manual changes

        if not self.users_auth_repo.is_git_repository_root:
            users_head_sha = None
        elif settings.overwrite_last_validated_commit:
            users_head_sha = last_validated_commit
        else:
            if last_validated_commit is not None:
                users_head_sha = self.users_auth_repo.top_commit_of_branch(
                    self.users_auth_repo.default_branch
                )
            else:
                # if the user's repository exists, but there is no last_validated_commit
                # start the update from the beginning
                users_head_sha = None

        if last_validated_commit != users_head_sha:
            # if a user committed something to the repo or manually pulled the changes
            # last_validated_commit will no longer match the top commit, but the repository
            # might still be completely valid
            # committing without pushing is not valid
            # user_head_sha should be newer than last validated commit
            if users_head_sha not in commits_since:
                msg = f"Top commit of repository {self.users_auth_repo.name} {users_head_sha} and is not equal to or newer than last successful commit"
                taf_logger.error(msg)
                raise UpdateFailedError(msg)
            users_head_sha = last_validated_commit

        # insert the current one at the beginning of the list
        if users_head_sha is not None:
            commits_since.insert(0, users_head_sha)

        self.commits = commits_since
        self.users_head_sha = users_head_sha
        self.current_commit_index = 0

    def _init_metadata(self):
        """
        TUF updater expects the existence of a client
        metadata directory. This directory stores
        the current metadata files (which will get deleted after the update).
        Directory must exist and contain at least root.json. Otherwise, update will
        fail. We actually want to validate the remote authentication repository,
        but will create a Temp directory in order to avoid modifying the updater.
        """
        metadata_files = self.validation_auth_repo.list_files_at_revision(
            self.current_commit, "metadata"
        )
        for filename in metadata_files:
            metadata = self.validation_auth_repo.get_file(
                self.current_commit, "metadata/" + filename
            )
            current_filename = self.metadata_dir_path / filename
            current_filename.write_text(metadata)

    def _patch_tuf_metadata_set(self, cls):
        """
        Used to address TUF metadata expiration.
        We skip validating expiration for metadata in older commits and
        revert-back to TUF metadata validation on the latest commit"""
        trusted_metadata_set.TrustedMetadataSet = cls

    def revert_tuf_patch_on_last_commit(f):
        """
        We only want to validate metadata expiration date for most recent commit.
        Since we turned off metadata validation for all commits, we want to revert this patch for the most recent commit.
        This decorator reverts our metadata expiration patch to original TUF implementation.
        """
        wraps(f)

        def wrapper(self):
            if self.current_commit_index + 1 == (len(self.commits) - 1):
                self._patch_tuf_metadata_set(self._original_tuf_trusted_metadata_set)
            return f(self)

        return wrapper

    def set_validation_repo(self, path, url):
        """
        Used outside of GitUpdater to access validation auth repo.
        """
        self.validation_auth_repo = AuthenticationRepository(path=path, urls=[url])

    def cleanup(self):
        """
        Removes the bare authentication repository and metadata
        directory. This should be called after the update is finished,
        either successfully or unsuccessfully.
        """
        if self.metadata_dir_path.is_dir():
            shutil.rmtree(self.metadata_dir_path)
        self.validation_auth_repo.cleanup()
        temp_dir = Path(self.validation_auth_repo.path, os.pardir).parent
        if temp_dir.is_dir():
            shutil.rmtree(str(temp_dir), onerror=on_rm_error)

    def get_current_targets(self):
        try:
            return self.validation_auth_repo.list_files_at_revision(
                self.current_commit, "targets"
            )
        except GitError:
            return []

    def get_current_target_data(self, filepath, raw=False):
        return self.validation_auth_repo.get_file(
            self.current_commit, f"targets/{filepath}", raw=raw
        )

    @revert_tuf_patch_on_last_commit
    def update_done(self):
        """Used to indicate whether updater has finished with update.
        Update is considered done when all commits have been validated"""
        self.current_commit_index += 1
        return self.current_commit_index == len(self.commits)



import enum
import glob
import json
import subprocess
import sys
from functools import partial
from pathlib import Path

import taf.settings as settings
from taf.repository_tool import get_target_path
from taf.utils import (
    run,
    safely_save_json_to_disk,
    extract_json_objects_from_trusted_stdout,
)
from taf.exceptions import GitError, ScriptExecutionError
from taf.log import taf_logger
from taf.updater.types.update import Update
from cattr import structure


class LifecycleStage(enum.Enum):
    REPO = "repo"
    UPDATE = "update"


class Event(enum.Enum):
    SUCCEEDED = "succeeded"
    CHANGED = "changed"
    UNCHANGED = "unchanged"
    FAILED = "failed"
    COMPLETED = "completed"


CONFIG_NAME = "config.json"
SCRIPTS_DIR = "scripts"
PERSISTENT_FILE_NAME = "persistent.json"


# config should only be read once per library root
# make it flexible in case the updater needs to handle libraries with different roots
config_db = {}


# persistent data should be read from persistent file and updated after every handler call
# should be one file per library root


def _get_script_path(lifecycle_stage, event):
    if settings.development_mode:
        return f"{lifecycle_stage.value}/{event.value}"
    else:
        return Path(
            get_target_path(f"{SCRIPTS_DIR}/{lifecycle_stage.value}/{event.value}")
        )


def get_config(library_root, config_name=CONFIG_NAME):
    global config_db
    if library_root not in config_db:
        try:
            config = json.loads(Path(library_root, CONFIG_NAME).read_text())
        except Exception:
            config = {}
        config_db[library_root] = config or {}
    return config_db.get(library_root)


def get_persistent_data(library_root, persistent_file=PERSISTENT_FILE_NAME):
    persistent_file = Path(library_root, PERSISTENT_FILE_NAME)
    if not persistent_file.is_file():
        persistent_file.touch()

    try:
        return json.loads(persistent_file.read_text())
    except Exception:
        return {}


def _handle_event(
    lifecycle_stage,
    event,
    transient_data,
    library_dir,
    scripts_root_dir,
    *args,
    **kwargs,
):
    # read initial persistent data from file
    taf_logger.info("Called {} handler. Event: {}", lifecycle_stage, event.value)
    persistent_data = get_persistent_data(library_dir)
    transient_data = transient_data or {}
    prepare_data_name = f"prepare_data_{lifecycle_stage.value}"
    # expect a dictionary containing a map of the authentication repository whose
    # scripts should be invoked and data to be passed to those scripts
    repos_and_data = globals()[prepare_data_name](
        event, transient_data, persistent_data, library_dir, *args, **kwargs
    )

    def _execute_scripts(repos_and_data, lifecycle_stage, event):
        scripts_rel_path = _get_script_path(lifecycle_stage, event)
        # this will update data
        for script_repo, script_data in repos_and_data.items():
            data = script_data["data"]
            last_commit = script_data["commit"]
            repos_and_data[script_repo]["data"] = execute_scripts(
                script_repo, last_commit, scripts_rel_path, data, scripts_root_dir
            )
        return repos_and_data

    if event in (Event.CHANGED, Event.UNCHANGED, Event.SUCCEEDED):
        # if event is changed or unchanged, execute these scripts first, then call the succeeded script
        if event == Event.CHANGED:
            repos_and_data = _execute_scripts(repos_and_data, lifecycle_stage, event)
        elif event == Event.UNCHANGED:
            repos_and_data = _execute_scripts(repos_and_data, lifecycle_stage, event)

        repos_and_data = _execute_scripts(
            repos_and_data, lifecycle_stage, Event.SUCCEEDED
        )
    elif event == Event.FAILED:
        repos_and_data = _execute_scripts(repos_and_data, lifecycle_stage, event)

    # execute completed handler at the end
    # _print_data(repos_and_data, library_dir, lifecycle_stage)
    repos_and_data = _execute_scripts(repos_and_data, lifecycle_stage, Event.COMPLETED)

    # return formatted response if update lifecycle is done
    if lifecycle_stage == LifecycleStage.UPDATE:
        for _, data in repos_and_data.items():
            data = data["data"]["update"]
            # example of converting data to classes. for now we're using cattrs default converters.
            # TODO: goal is full support of class types
            return structure(data, Update)
    # return transient data as it should be propagated to other event and handlers
    return {
        repo.name: data["data"]["state"]["transient"]
        for repo, data in repos_and_data.items()
    }


handle_repo_event = partial(_handle_event, LifecycleStage.REPO)
handle_update_event = partial(_handle_event, LifecycleStage.UPDATE)


def execute_scripts(auth_repo, last_commit, scripts_rel_path, data, scripts_root_dir):
    persistent_path = auth_repo.library_dir / PERSISTENT_FILE_NAME
    # do not load the script from the file system
    # the update might have failed because the repository contains an additional
    # commit with, say, malicious scripts
    # load the scripts from the last validated commit - whether new or old

    # this is a nightmare to test
    # load from filesystem in development mode so that the scripts can be updated without
    # having to commit and push
    development_mode = settings.development_mode

    if development_mode:
        if scripts_root_dir is not None:
            path = Path(scripts_root_dir) / auth_repo.name / scripts_rel_path
        else:
            path = Path(auth_repo.path) / scripts_rel_path
        script_paths = glob.glob(f"{path}/*.py")
    else:
        try:
            script_names = auth_repo.list_files_at_revision(
                last_commit, scripts_rel_path
            )
            script_rel_paths = [
                (scripts_rel_path / script_name).as_posix()
                for script_name in script_names
                if Path(script_name).suffix == ".py"
            ]
            auth_repo.checkout_paths(last_commit, *script_rel_paths)
            script_paths = [
                str(auth_repo.path / script_rel_path)
                for script_rel_path in script_rel_paths
            ]
        except GitError:
            taf_logger.debug(f"{scripts_rel_path} does not contain any scripts")
            script_paths = []

    for script_path in sorted(script_paths):
        taf_logger.info("Executing script {}", script_path)
        json_data = json.dumps(data)
        try:
            output = run(sys.executable, script_path, input=json_data)
        except subprocess.CalledProcessError as e:
            taf_logger.error(
                "An error occurred while executing {}: {}", script_path, e.output
            )
            raise ScriptExecutionError(script_path, e.output)
        if output is not None and output != "":
            # if the script contains print statements other than the final
            # print which outputs transient and persistent data
            # meaning that we might not be able to convert output to a json
            # so try to locate jsons inside the output
            for json_data in extract_json_objects_from_trusted_stdout(output):
                transient_data = json_data.get("transient")
                persistent_data = json_data.get("persistent")
                if transient_data is not None or persistent_data is not None:
                    break
        else:
            transient_data = persistent_data = {}
        taf_logger.debug("Persistent data: {}", persistent_data)
        taf_logger.debug("Transient data: {}", transient_data)
        # overwrite current persistent and transient data
        data["state"]["transient"] = transient_data
        data["state"]["persistent"] = persistent_data
        try:
            # if persistent data is not a valid json or if an error happens while storing
            # to disk, raise an error
            # we save to disk by copying a temp file, so the likelihood of an error
            # should be decreased
            safely_save_json_to_disk(persistent_data, persistent_path)
        except Exception as e:
            raise ScriptExecutionError(
                script_path,
                f"An error occurred while saving persistent data to disk: {str(e)}",
            )
    return data


def get_script_repo_and_commit_repo(auth_repo, commits_data, *args):
    return auth_repo, commits_data["after_pull"]


def get_script_repo_and_commit_update(auth_repo, commits_data, *args):
    pass


def prepare_data_repo(
    event,
    transient_data,
    persistent_data,
    library_dir,
    auth_repo,
    commits_data,
    error,
    targets_data,
):
    return {
        auth_repo: {
            "data": {
                "update": _repo_update_data(
                    auth_repo, event, commits_data, targets_data, error
                ),
                "state": {
                    "transient": transient_data,
                    "persistent": persistent_data,
                },
                "config": get_config(auth_repo.library_dir),
            },
            "commit": commits_data["after_pull"],
        }
    }


def prepare_data_update(
    event,
    transient_data,
    persistent_data,
    library_dir,
    repos_update_data,
    error,
    root_auth_repo,
):

    all_auth_repos = {}
    for auth_repo_name in repos_update_data:
        all_auth_repos[auth_repo_name] = _repo_update_data(
            **repos_update_data[auth_repo_name]
        )

    return {
        root_auth_repo: {
            "data": {
                "update": {
                    "changed": event == Event.CHANGED,
                    "event": _format_event(event),
                    "error_msg": str(error) if error else "",
                    "auth_repos": all_auth_repos,
                    "auth_repo_name": root_auth_repo.name,
                },
                "state": {
                    "transient": transient_data,
                    "persistent": persistent_data,
                },
                "config": get_config(library_dir),
            },
            "commit": repos_update_data[root_auth_repo.name]["commits_data"][
                "after_pull"
            ],
        }
    }


def _repo_update_data(auth_repo, update_status, commits_data, targets_data, error):
    return {
        "changed": update_status == Event.CHANGED,
        "event": _format_event(update_status),
        "repo_name": auth_repo.name,
        "error_msg": str(error) if error else "",
        "auth_repo": {
            "data": auth_repo.to_json_dict(),
            "commits": commits_data,
        },
        "target_repos": targets_data,
    }


def _print_data(repos_and_data, library_dir, lifecycle_stage):
    for repo, data in repos_and_data.items():
        path = Path(library_dir, "data", lifecycle_stage.value, f"{repo.name}.json")
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(data, indent=4))


def _format_event(event):
    if event in (Event.CHANGED, Event.UNCHANGED, Event.SUCCEEDED):
        return f"event/{Event.SUCCEEDED.value}"
    else:
        return f"event/{Event.FAILED.value}"



from getpass import getpass
from os import getcwd
from pathlib import Path
from typing import Dict, Optional

import click
import securesystemslib
from securesystemslib.interface import (
    import_rsa_privatekey_from_file,
    import_rsa_publickey_from_file,
)
from taf.repository_tool import Repository
from tuf.repository_tool import import_rsakey_from_pem

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import KeystoreError
from taf.log import taf_logger


def default_keystore_path() -> str:
    keystore_path = str(Path(getcwd(), "keystore"))
    taf_logger.info(f"Keystore path is not passed, using default one: {keystore_path}")
    return keystore_path


def _form_private_pem(pem: str) -> str:
    return f"-----BEGIN RSA PRIVATE KEY-----\n{pem}\n-----END RSA PRIVATE KEY-----"


def _from_public_pem(pem: str) -> str:
    return f"-----BEGIN PUBLIC KEY-----\n{pem}\n-----END PUBLIC KEY-----"


def key_cmd_prompt(
    key_name: str,
    role: str,
    taf_repo: Repository,
    loaded_keys: Optional[Dict] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
) -> Optional[Dict]:
    def _enter_and_check_key(key_name, role, loaded_keys, scheme):
        pem = getpass(f"Enter {key_name} private key without its header and footer\n")
        pem = _form_private_pem(pem)
        try:
            key = import_rsakey_from_pem(pem, scheme)
        except Exception:
            print("Invalid key")
            return None
        public_key = import_rsakey_from_pem(key["keyval"]["public"], scheme)
        if not taf_repo.is_valid_metadata_yubikey(role, public_key):
            print(f"The entered key is not a valid {role} key")
            return None
        if loaded_keys is not None and key in loaded_keys:
            print("Key already entered")
            return None
        return key

    while True:
        key = _enter_and_check_key(key_name, role, loaded_keys, scheme)
        if key is not None:
            return key


def load_tuf_private_key(
    key_str: str, key_name: str, scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME
) -> Dict:
    if not key_str:
        key_str = getpass(
            f"Enter {key_name} private key without its header and footer\n"
        )
    key_pem = _form_private_pem(key_str)

    return import_rsakey_from_pem(key_pem, scheme)


def new_public_key_cmd_prompt(scheme: str) -> Dict:
    def _enter_and_check_key(scheme):
        pem = getpass("Enter public key without its header and footer\n")
        pem = _from_public_pem(pem)
        try:
            key = import_rsakey_from_pem(pem, scheme)
        except Exception:
            print("Invalid key")
            return None
        return import_rsakey_from_pem(key["keyval"]["public"], scheme)

    while True:
        key = _enter_and_check_key(scheme)
        if key is not None:
            return key


def read_private_key_from_keystore(
    keystore: str,
    key_name: str,
    key_num: Optional[int] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    password: Optional[str] = None,
) -> Dict:
    key_path = Path(keystore, key_name).expanduser().resolve()
    if not key_path.is_file():
        raise KeystoreError(f"{str(key_path)} does not exist")

    def _read_key(path, password, scheme):
        def _read_key_or_keystore_error(path, password, scheme):
            try:
                return import_rsa_privatekey_from_file(
                    str(Path(keystore, key_name)), password or None, scheme=scheme
                )
            except (
                securesystemslib.exceptions.FormatError,
                securesystemslib.exceptions.Error,
            ) as e:
                if "password" in str(e).lower():
                    return None
                raise KeystoreError(e)

        try:
            # try to load with a given password or None
            return _read_key_or_keystore_error(path, password, scheme)
        except securesystemslib.exceptions.CryptoError:
            password = getpass(
                f"Enter {key_name} keystore file password and press ENTER"
            )
            return _read_key_or_keystore_error(path, password, scheme)
        except Exception:
            return None

    while True:
        key = _read_key(key_path, password, scheme)
        if key is not None:
            return key
        if not click.confirm("Could not open keystore file. Try again?"):
            raise KeystoreError(f"Could not open keystore file {key_path}")


def read_public_key_from_keystore(
    keystore: str, key_name: str, scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME
) -> Dict:
    pub_key_path = Path(keystore, f"{key_name}.pub").expanduser().resolve()
    if not pub_key_path.is_file():
        raise KeystoreError(f"{str(pub_key_path)} does not exist")
    try:
        return import_rsa_publickey_from_file(str(pub_key_path), scheme)
    except (
        securesystemslib.exceptions.FormatError,
        securesystemslib.exceptions.Error,
    ) as e:
        raise KeystoreError(e)



from __future__ import annotations
import json
import itertools
import os
import re
import pygit2
import subprocess
import logging
from collections import OrderedDict
from functools import reduce
from pathlib import Path

import taf.settings as settings
from taf.exceptions import (
    TAFError,
    CloneRepoException,
    FetchException,
    InvalidRepositoryError,
    GitError,
)
from taf.log import taf_logger
from taf.utils import run
from typing import Callable, Dict, List, Optional, Tuple
from .pygit import PyGitRepository

EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"


class GitRepository:
    def __init__(
        self,
        library_dir: Optional[str] = None,
        name: Optional[str] = None,
        urls: Optional[List[str]] = None,
        custom: Optional[Dict] = None,
        default_branch: Optional[str] = None,
        allow_unsafe: Optional[bool] = False,
        path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        Args:
          library_dir (Path or str): path to the library root. This is a directory which contains other repositories,
          whose full path is determined by appending their names to the library dir. A repository can be
          instantiated by specifying library dir and name, or by the full path.
          name (str): repository's name, which is appended to the library dir to form the full path.
          Must be in the namespace/name format. If library_dir is specified, name must be specified too.
          path (Path): repository's full filesystem path, which can be specified instead of name and library dir
          urls (list): repository's urls
          custom (dict): a dictionary containing other data
          default_branch (str): repository's default branch, automatically determined if not specified
          allow_unsafe: allow a git's security mechanism which prevents execution of git commands if
          the containing directory is owned by a different user to be ignored
        """
        if isinstance(library_dir, str):
            library_dir = Path(library_dir)
        if isinstance(path, str):
            path = Path(path)

        if (library_dir, name).count(None) == 1:
            raise InvalidRepositoryError(
                "Both library_dir and name need to be specified"
            )

        if name is not None:
            self.name = self._validate_repo_name(name)
            self.path = self._validate_repo_path(library_dir, name, path)
            self.library_dir = library_dir.expanduser().resolve()
        elif path is None:
            raise InvalidRepositoryError(
                "Either specify library dir and name pair or path!"
            )
        else:
            # maintain support for repositories whose names are not of significance
            # in that case, only full path is specified (can happen if only using the GitRepository class)
            # without the rest of the framework in some cotext
            # name is still required for logging, so determine it based on the path
            # if the path points to a direcotry directly inside the root direcotry
            # set name to the name of that folder
            # otherwise, use the same format that is expected when name is specified
            if path.parent.parent != path.parent:
                self.name = f"{path.parent.name}/{path.name}"
                self.library_dir = path.parent.parent
            else:
                self.name = path.name
                self.library_dir = path.parent
            self.library_dir = self.library_dir.resolve()
            self.path = self._validate_repo_path(path, None)

        self.urls = self._validate_urls(urls)
        self.allow_unsafe = allow_unsafe
        self.custom = custom or {}
        if default_branch is None:
            default_branch = self._determine_default_branch()
        self.default_branch = default_branch

    _pygit = None

    @property
    def pygit(self):
        if self._pygit is None:
            try:
                self._pygit = PyGitRepository(self)
            except Exception:
                pass
        return self._pygit

    @classmethod
    def from_json_dict(cls, json_data: Dict):
        """Create a new instance based on data contained by the `json_data` dictionary,
        which can be create by calling `to_json_dict`
        """
        return cls(**json_data)

    def to_json_dict(self):
        """Returns a dictionary mapping all attributes to their values"""
        return {
            "library_dir": str(self.library_dir),
            "name": self.name,
            "urls": self.urls,
            "default_branch": self.default_branch,
            "custom": self.custom,
        }

    logging_functions = {
        logging.DEBUG: taf_logger.debug,
        logging.INFO: taf_logger.info,
        logging.WARNING: taf_logger.warning,
        logging.ERROR: taf_logger.error,
        logging.CRITICAL: taf_logger.critical,
    }

    log_template = "{}{}"

    _remotes = None

    @property
    def remotes(self) -> List[str]:
        if self._remotes is None:
            repo = self.pygit_repo
            self._remotes = [remote.name for remote in repo.remotes]
        return self._remotes

    @property
    def is_git_repository_root(self) -> bool:
        """Check if path is git repository."""
        git_path = self.path / ".git"
        return self.is_git_repository and (git_path.is_dir() or git_path.is_file())

    @property
    def is_git_repository(self) -> bool:
        repo_path = pygit2.discover_repository(str(self.path))
        return repo_path is not None

    @property
    def initial_commit(self) -> str:
        return (
            self._git(
                "rev-list --max-parents=0 HEAD", error_if_not_exists=False
            ).strip()
            if self.is_git_repository
            else None
        )

    @property
    def log_prefix(self) -> str:
        return f"Repo {self.name}: "

    @property
    def pygit_repo(self) -> Optional[pygit2.Repository]:
        try:
            return self.pygit.repo
        except Exception as e:
            self._log_debug(f"Unable to instantiate pygit2 repo due to error: {str(e)}")
            return None

    def _git(self, cmd, *args, **kwargs):
        """Call git commands in subprocess
        e.g.:
          self._git('checkout {}', branch_name)
        """
        log_error = kwargs.pop("log_error", False)
        log_error_msg = kwargs.pop("log_error_msg", "")
        reraise_error = kwargs.pop("reraise_error", False)
        log_success_msg = kwargs.pop("log_success_msg", "")
        error_if_not_exists = kwargs.pop("error_if_not_exists", True)

        if len(args):
            cmd = cmd.format(*args)
        if self.allow_unsafe:
            command = f"git -C {self.path} -c safe.directory={self.path} {cmd}"
        else:
            command = f"git -C {self.path} {cmd}"
        result = None
        if log_error or log_error_msg:
            try:
                result = run(command, **kwargs)
                if log_success_msg:
                    self._log_debug(log_success_msg)
            except subprocess.CalledProcessError as e:
                if error_if_not_exists and (
                    not self.path.is_dir() or not self.is_git_repository
                ):
                    log_error_msg = (
                        f"{self.path} does not exist or is not a git repository"
                    )
                    reraise_error = True
                if log_error_msg:
                    error = GitError(self, message=log_error_msg, error=e)
                    self._log_error(error.message)
                else:
                    error = GitError(self, command=command, error=e)
                    # not every git error indicates a problem
                    # if it does, we expect that either custom error message will be provided
                    # or that the error will be reraised
                    self._log_debug(error.message)
                if reraise_error:
                    raise error
        else:
            try:
                result = run(command, **kwargs)
            except subprocess.CalledProcessError as e:
                raise GitError(self, command=command, error=e)
            if log_success_msg:
                self._log_debug(log_success_msg)
        return result

    def _get_default_branch_from_local(self) -> str:
        try:
            branch = self._git(
                "symbolic-ref refs/remotes/origin/HEAD --short", reraise_error=True
            )
            return _local_branch_re.sub("", branch)
        except GitError as e:
            self._log_debug(f"Could not get remote HEAD at {self.path}: {e}")
            pass
        # NOTE: will not work if local repository is in a detached HEAD
        branch = self._git("symbolic-ref HEAD --short", reraise_error=True)
        return branch

    def _get_default_branch_from_remote(self, url: str) -> str:
        branch = self._git(
            f"ls-remote --symref {url} HEAD",
            log_error=True,
            log_error_msg="Unable to get default branch from remote",
            reraise_error=True,
        )
        branch = branch.split("\t", 1)[0]
        branch = branch.split()[-1]
        return _remote_branch_re.sub("", branch)

    def _log(self, log_func: Callable, message: str) -> None:
        log_func(self.log_template, self.log_prefix, message)

    def _log_debug(self, message: str) -> None:
        self._log(self.logging_functions[logging.DEBUG], message)

    def _log_info(self, message: str) -> None:
        self._log(self.logging_functions[logging.INFO], message)

    def _log_warning(self, message: str) -> None:
        self._log(self.logging_functions[logging.WARNING], message)

    def _log_error(self, message: str) -> None:
        self._log(self.logging_functions[logging.ERROR], message)

    def _log_critical(self, message: str) -> None:
        self._log(self.logging_functions[logging.CRITICAL], message)

    def all_commits_on_branch(
        self, branch: Optional[str] = None, reverse: Optional[bool] = True
    ) -> List[str]:
        """Returns a list of all commits on the specified branch.
        If branch is None, all commits on the currently checked out branch will be returned
        """
        repo = self.pygit_repo
        if branch:
            branch = repo.branches.get(branch)
            if branch is None:
                raise GitError(
                    self,
                    message=f"Error occurred while getting commits of branch {branch}. Branch does not exist",
                )
            latest_commit_id = branch.target
        else:
            latest_commit_id = repo[repo.head.target].id

        sort = pygit2.GIT_SORT_REVERSE if reverse else pygit2.GIT_SORT_NONE
        commits = [commit.id.hex for commit in repo.walk(latest_commit_id, sort)]
        self._log_debug(f"found the following commits: {', '.join(commits)}")
        return commits

    def all_commits_since_commit(
        self,
        since_commit: str,
        branch: Optional[str] = None,
        reverse: Optional[bool] = True,
    ) -> List[str]:
        """Returns a list of all commits since the specified commit on the
        specified or currently checked out branch

        Raises:
            exceptions.GitError: An error occured with provided commit SHA
        """
        if since_commit is None:
            return self.all_commits_on_branch(branch=branch, reverse=reverse)

        try:
            self.commit_exists(commit_sha=since_commit)
        except GitError as e:
            self._log_warning(f"Commit {since_commit} not found in local repository.")
            raise e
        repo = self.pygit_repo
        if branch:
            branch = repo.branches.get(branch)
            latest_commit_id = branch.target
        else:
            latest_commit_id = repo[repo.head.target].id

        shas = []
        for commit in repo.walk(latest_commit_id):
            sha = commit.id.hex
            if sha == since_commit:
                break
            shas.insert(0, sha)
        if not reverse:
            shas = shas[::-1]
        self._log_debug(f"found the following commits: {', '.join(shas)}")
        return shas

    def add_remote(self, upstream_name: str, upstream_url: str) -> None:
        try:
            self._git("remote add {} {}", upstream_name, upstream_url)
        except GitError as e:
            if "already exists" not in str(e):
                raise

    def branches(
        self, remote: bool = False, all: bool = False, strip_remote: bool = False
    ) -> List[str]:
        """Returns all branches."""
        repo = self.pygit_repo
        if all:
            branches = list(repo.branches)
        elif remote:
            branches = list(repo.branches.remote)
        else:
            branches = list(repo.branches.local)

        if strip_remote:
            remotes = self.remotes
            branches = set(
                [
                    reduce(lambda b, r: b.replace(f"{r}/", ""), remotes, branch)
                    for branch in branches
                ]
            )

        return branches

    def branches_containing_commit(
        self,
        commit: str,
        strip_remote: Optional[bool] = False,
        sort_key: Optional[Callable] = None,
    ) -> OrderedDict:
        """Finds all branches that contain the given commit"""
        repo = self.pygit_repo
        local_branches = remote_branches = []
        try:
            local_branches = list(repo.branches.local.with_commit(commit))
        except pygit2.GitError:
            pass
        try:
            remote_branches = list(repo.branches.remote.with_commit(commit))
        except pygit2.GitError:
            pass
        filtered_remote_branches = []
        if len(remote_branches):
            for branch in remote_branches:
                local_name = self.branch_local_name(branch)
                if (
                    local_name
                    and "HEAD" not in local_name
                    and local_name not in local_branches
                ):
                    if strip_remote:
                        branch = self.branch_local_name(branch)
                    filtered_remote_branches.append(branch)
        branches = {branch: False for branch in local_branches}
        branches.update({branch: True for branch in filtered_remote_branches})
        return OrderedDict(sorted(branches.items(), key=sort_key, reverse=True))

    def branch_exists(
        self, branch_name: str, include_remotes: Optional[bool] = True
    ) -> bool:
        """
        Checks if a branch with the given name exists.
        If include_remotes is set to True, this checks if
        a remote branch exists.
        """
        repo = self.pygit_repo
        branch = repo.branches.get(branch_name)
        # this git command should return the branch's name if it exists
        # empty string otherwise
        if branch is not None:
            return True
        if include_remotes:
            for remote in self.remotes:
                remote_branch = repo.branches.remote.get(f"{remote}/{branch_name}")
                if remote_branch is not None:
                    return remote_branch

                # finally, check remote branch
                if self.has_remote():
                    return branch_name in self._git(
                        f"ls-remote --heads origin {branch_name}",
                        log_error_msg=f"Repo {self.name}: could check if the branch exists in the remote repository",
                        reraise_error=True,
                    )

        return False

    def branch_off_commit(self, branch_name: str, commit_sha: str) -> None:
        """Create a new branch by branching off of the specified commit"""
        repo = self.pygit_repo
        try:
            commit = repo[commit_sha]
            repo.branches.local.create(branch_name, commit)
            branch = repo.lookup_branch(branch_name)
            ref = repo.lookup_reference(branch.name)
            repo.checkout(ref)
        except Exception as e:
            self._log_error(str(e))
            raise
        finally:
            self._log_info(
                f"Created a new branch {branch_name} from branching off of {commit_sha}"
            )

    def branch_local_name(self, remote_branch_name: str) -> str:
        """Strip remote from the given remote branch"""
        for remote in self.remotes:
            if remote_branch_name.startswith(remote + "/"):
                return remote_branch_name.split("/", 1)[1]

    def checkout_branch(
        self,
        branch_name: str,
        create: Optional[bool] = False,
        raise_anyway: bool = False,
    ) -> None:
        """Check out the specified branch. If it does not exists and
        the create parameter is set to True, create a new branch.
        If the branch does not exist and create is set to False,
        raise an exception."""
        repo = self.pygit_repo
        try:
            branch = repo.lookup_branch(branch_name)
            if branch is not None:
                ref = repo.lookup_reference(branch.name)
                repo.checkout(ref)
            else:
                self._git(
                    "checkout {}",
                    branch_name,
                    log_success_msg=f"Repo {self.name}: checked out branch {branch_name}",
                )
        except Exception as e:
            if raise_anyway:
                raise GitError(repo=self, message=str(e))
            # skip worktree errors
            if "the current HEAD of a linked repository" in str(e):
                return

            if create:
                self.create_and_checkout_branch(branch_name)
            else:
                self._log_error(f"could not checkout branch {branch_name}")
                raise (e)

    def check_if_clean_and_synced(self, branch: Optional[str] = None) -> bool:
        """
        Check if repository's worktree is clean and if the
        specified (or default) branch is synced with remote
        (if the repository has a remote set)
        """
        branch = branch or self.default_branch
        if self.something_to_commit():
            return False
        if not self.has_remote():
            return True
        if not self.synced_with_remote(branch):
            return False

    def checkout_paths(self, commit_sha: str, *args) -> None:
        repo = self.pygit_repo
        commit = repo.get(commit_sha)
        repo.checkout_tree(commit, paths=list(args))

    def checkout_orphan_branch(self, branch_name: str) -> None:
        """Creates orphan branch"""
        self._git(
            f"checkout --orphan {branch_name}", log_error=True, reraise_error=True
        )
        try:
            self._git("rm -rf .")
        except GitError:  # If repository is empty
            pass

    def clean(self):
        self._git("clean -fd")

    def cleanup(self):
        if self._pygit is not None:
            self._pygit.cleanup()
            self._pygit = None

    def clone(
        self, no_checkout: bool = False, bare: Optional[bool] = False, **kwargs
    ) -> None:
        self._log_info("cloning repository")

        self.path.mkdir(exist_ok=True, parents=True)
        if len(os.listdir(self.path)) != 0:
            raise GitError(
                repo=self,
                message=f"destination path {self.path} is not an empty directory.",
            )

        if self.urls is None:
            raise GitError(
                repo=self, message="cannot clone repository. No urls were specified"
            )
        params = []
        if bare:
            params.append("--bare")
        elif no_checkout:
            params.append("--no-checkout")

        for name, value in kwargs.items():
            if isinstance(value, bool):
                params.append(f"--{name}")
            else:
                params.append(f"--{name} {value}")

        params = " ".join(params)

        cloned = False
        for url in self.urls:
            self._log_info(f"trying to clone from {url}")
            try:
                self._git(
                    "clone {} . {}",
                    url,
                    params,
                    log_success_msg=f"successfully cloned from {url}",
                    log_error_msg=f"cannot clone from url {url}",
                    reraise_error=True,
                )
            except GitError as e:
                self._log_info(f"could not clone from {url} due to {e}")
            else:
                self._log_info(f"successfully cloned from {url}")
                cloned = True
                break

        if not cloned:
            raise CloneRepoException(self)

        if self.default_branch is None:
            self.default_branch = self._determine_default_branch()

    def clone_or_pull(
        self,
        branches: Optional[List[str]] = None,
        only_fetch: Optional[bool] = False,
        **kwargs,
    ) -> Tuple[str, str]:
        """
        Clone or fetch the specified branch for the given repo.
        Return old and new HEAD.
        """

        old_head = self.head_commit_sha()
        if old_head is None:
            self._log_debug(f"cloning {self.name}")
            self._log_debug(f"old head sha is {old_head}")
            self.clone(**kwargs)
        else:
            if branches is None:
                branches = [self.default_branch]
            self._log_debug(f"pulling branches {', '.join(branches)}")
            try:
                for branch in branches:
                    if only_fetch:
                        self._git("fetch", "origin", f"{branch}:{branch}")
                    else:
                        self._git("pull", "origin", branch)
                    self._log_info(f"successfully fetched branch {branch}")
            except GitError as e:
                if "fatal" in str(e):
                    raise FetchException(f"{self.path}: {str(e)}")
                pass

        new_head = self.head_commit_sha()

        return old_head, new_head

    def create_and_checkout_branch(
        self, branch_name: str, raise_error_if_exists: Optional[bool] = True
    ) -> None:
        repo = self.pygit_repo
        try:
            branch = repo.lookup_branch(branch_name)
            if branch is not None and raise_error_if_exists:
                raise GitError(f"Branch {branch_name} already exists")
            try:
                commit = repo.revparse_single("HEAD")
                repo.branches.local.create(branch_name, commit)
                branch = repo.lookup_branch(branch_name)
                ref = repo.lookup_reference(branch.name)
                repo.checkout(ref)
            except KeyError:
                # this will be execute if there is no HEAD pointer
                flag = "-b" if raise_error_if_exists else "-B"
                self._git(
                    "checkout {} {}",
                    flag,
                    branch_name,
                    log_success_msg=f"created and checked out branch {branch_name}",
                    log_error=True,
                    reraise_error=True,
                )

        except Exception as e:
            self._log_error(str(e))
            raise
        finally:
            self._log_info(f"Created a new branch {branch_name}")

    def create_branch(self, branch_name: str, commit: Optional[str] = None) -> None:
        repo = self.pygit_repo
        try:
            if commit is not None:
                branch_commit = repo[commit]
            else:
                branch_commit = repo.revparse_single("HEAD")
            repo.branches.local.create(branch_name, branch_commit)
        except Exception as e:
            self._log_error(str(e))
            raise
        finally:
            self._log_info(f"Created a new branch {branch_name}")

    def checkout_commit(self, commit: str) -> None:
        self._git(
            "checkout {}",
            commit,
            log_success_msg=f"checked out commit {commit}",
            log_error=True,
            reraise_error=True,
        )

    def commit(self, message: str) -> None:
        """Create a commit with the provided message on the currently checked out branch"""
        # This implementation is faster than with pygit2 because adding files to index is
        # significantly faster with `git add -A`
        self._git("add -A")
        try:
            self._git("diff --cached --exit-code --shortstat", reraise_error=True)
        except GitError:
            try:
                run("git", "-C", str(self.path), "commit", "--quiet", "-m", message)
            except subprocess.CalledProcessError as e:
                raise GitError(
                    repo=self, message=f"could not commit changes due to:\n{e}"
                )
        return self._git("rev-parse HEAD")

    def commit_empty(self, message: str) -> None:
        run(
            "git",
            "-C",
            str(self.path),
            "commit",
            "--quiet",
            "--allow-empty",
            "-m",
            message,
        )
        return self._git("rev-parse HEAD")

    def commit_exists(self, commit_sha: str) -> str:
        return self._git(f"rev-parse {commit_sha}")

    def commits_on_branch_and_not_other(self, branch1: str, branch2: str) -> List[str]:
        """
        Meant to find commits belonging to a branch which branches off of
        a commit from another branch. For example, to find only commits
        on a speculative branch and not on the main branch.
        """
        merge_base = self.get_merge_base(branch1, branch2)
        commits = self.all_commits_since_commit(merge_base, branch1, reverse=False)

        return commits

    def commit_before_commit(self, commit: str) -> Optional[str]:
        repo = self.pygit_repo
        repo_commit_id = repo.get(commit).id
        for comm in repo.walk(repo_commit_id):
            hex = comm.id.hex
            if hex != commit:
                return hex
        return None

    def delete_local_branch(self, branch_name: str) -> None:
        """Deletes local branch."""
        try:
            repo = self.pygit_repo
            repo.branches.delete(branch_name)
        except KeyError:
            raise GitError(
                repo=self,
                message=f"Could not delete branch {branch_name}. Branch does not exist",
            )
        except Exception:
            raise GitError(repo=self, message=f"Could not delete branch {branch_name}.")

    def delete_remote_tracking_branch(
        self, remote_branch_name: str, force: Optional[bool] = False
    ) -> None:
        flag = "-D" if force else "-d"
        self._git(f"branch {flag} -r {remote_branch_name}", log_error=True)

    def delete_remote_branch(
        self, branch_name: str, remote: Optional[bool] = None
    ) -> None:
        if remote is None:
            remote = self.remotes[0]
        self._git(f"push {remote} --delete {branch_name}", log_error=True)

    def get_commit_message(self, commit_sha: str) -> str:
        """Returns commit message of the given commit"""
        repo = self.pygit_repo
        commit = repo.get(commit_sha)
        return commit.message

    def get_commit_sha(self, behind_head: str) -> str:
        """Get commit sha of HEAD~{behind_head}"""
        return self._git("rev-parse HEAD~{}", behind_head)

    def get_default_branch(self, url: Optional[str] = None) -> str:
        """Get the default branch of the repository. If url is provided, return the
        default branch from the remote. Otherwise, return the default
        branch from the local repository."""
        if url is not None:
            url = url.strip()
            return self._get_default_branch_from_remote(url)
        return self._get_default_branch_from_local()

    def get_json(self, commit: str, path: str, raw: Optional[bool] = False) -> Dict:
        s = self.get_file(commit, path, raw=raw)
        return json.loads(s)

    def get_file(
        self,
        commit: str,
        path: str,
        raw: Optional[bool] = False,
        with_id: Optional[bool] = False,
    ) -> Tuple[str, str] | str:
        path = Path(path).as_posix()
        try:
            git_id, content = self.pygit.get_file(commit, path, raw)
            if with_id:
                return git_id, content
            return content
        except TAFError as e:
            raise e
        except Exception:
            return self._git("show {}:{}", commit, path, raw=raw)

    def get_first_commit_on_branch(self, branch: Optional[str] = None) -> str:
        branch = branch or self.default_branch
        first_commit = self._git(
            f"rev-list --max-parents=0 {branch}", error_if_not_exists=False
        )
        return first_commit.strip() if first_commit else None

    def get_last_branch_by_committer_date(self) -> Optional[str]:
        """Find the latest branch based on committer date. Should only be used for
        testing purposes"""
        branches = self._git("branch --sort=committerdate").strip().split("\n")
        if not len(branches):
            return None
        return branches[-1]

    def get_remote_url(self) -> Optional[str]:
        try:
            return self._git("config --get remote.origin.url").strip()
        except GitError:
            return None

    def delete_branch(self, branch_name: str) -> None:
        self._git("branch -D {}", branch_name, log_error=True)

    def diff_between_revisions(
        self, revision1: Optional[str] = EMPTY_TREE, revision2: Optional[str] = "HEAD"
    ) -> str:
        return self._git("diff --name-status {} {}", revision1, revision2)

    def has_remote(self) -> bool:
        return len(self.remotes) > 0

    def head_commit_sha(self) -> Optional[str]:
        """Finds sha of the commit to which the current HEAD points"""
        repo = self.pygit_repo
        try:
            return repo.revparse_single("HEAD").id.hex
        except Exception:
            return None

    def fetch(
        self,
        fetch_all: Optional[bool] = False,
        branch: Optional[str] = None,
        remote: Optional[str] = "origin",
    ) -> None:
        if fetch_all:
            self._git("fetch --all", log_error=True)
        else:
            if branch is None:
                branch = ""
            self._git("fetch {} {}", remote, branch, log_error=True)

    def find_worktree_path_by_branch(self, branch_name: str) -> Optional[str]:
        """Returns path of the workree where the branch is checked out, or None if not checked out in any worktree"""
        worktrees = self.list_worktrees()
        for path, _, _branch_name in worktrees.values():
            if _branch_name == branch_name:
                return path
        return None

    def get_current_branch(self, full_name: Optional[str] = False) -> str:
        """Return current branch."""
        repo = self.pygit_repo
        branch = repo.lookup_reference("HEAD").resolve()
        if full_name:
            return branch.name
        return branch.shorthand

    def get_last_remote_commit(self, url: str, branch: Optional[str] = None) -> str:
        """
        Fetch the last remote commit of the specified branch
        """
        branch = branch or self.default_branch
        if url is None:
            raise FetchException(
                "Could not fetch the last remote commit. URL not specified"
            )
        last_commit = self._git(f"--no-pager ls-remote {url} {branch}", log_error=True)
        if last_commit:
            last_commit = last_commit.split("\t", 1)[0]
            # in some cases (e.g. upstream is defined the result might contain a warning line)
            return last_commit.split()[-1]

    def get_merge_base(self, branch1: str, branch2: str) -> str:
        """Finds the best common ancestor between two branches"""
        repo = self.pygit_repo
        commit1 = self.top_commit_of_branch(branch1)
        commit2 = self.top_commit_of_branch(branch2)
        return repo.merge_base(commit1, commit2).hex

    def get_tracking_branch(
        self, branch: Optional[str] = "", strip_remote: Optional[bool] = False
    ) -> Optional[str]:
        """Returns tracking branch name in format origin/branch-name or None if branch does not
        track remote branch.
        """
        try:
            tracking_branch = self._git(f"rev-parse --abbrev-ref {branch}@{{u}}")
            if strip_remote:
                tracking_branch = self.branch_local_name(tracking_branch)
            return tracking_branch
        except GitError:
            return None

    def init_repo(self, bare: Optional[bool] = False) -> None:
        if self.path.is_dir():
            self.path.mkdir(exist_ok=True, parents=True)
        flag = "--bare" if bare else ""
        self._git(f"init {flag}", error_if_not_exists=False)
        if self.urls is not None and len(self.urls):
            self._git("remote add origin {}", self.urls[0])

    def is_remote_branch(self, branch_name: str) -> bool:
        for remote in self.remotes:
            if branch_name.startswith(remote + "/"):
                return True
        return False

    def is_bare_repository(self) -> bool:
        return self.pygit_repo.is_bare

    def list_files_at_revision(
        self, commit: str, path: Optional[str] = ""
    ) -> List[str]:
        path = Path(path).as_posix()
        try:
            return self.pygit.list_files_at_revision(commit, path)
        except TAFError as e:
            raise e
        except Exception:
            self._log_warning(
                "Perfomance regression: Could not list files with pygit2. Reverting to git subprocess"
            )
            return self._list_files_at_revision(commit, path)

    def _list_files_at_revision(self, commit: str, path: str) -> List[str]:
        if path is None:
            path = ""
        file_names = self._git("ls-tree -r --name-only {}", commit)
        list_of_files = []
        if not file_names:
            return list_of_files
        for file_in_repo in file_names.split("\n"):
            if not file_in_repo.startswith(path):
                continue
            file_in_repo = os.path.relpath(file_in_repo, path)
            list_of_files.append(file_in_repo)
        return list_of_files

    def list_changed_files_at_revision(self, commit: str) -> List[str]:
        repo = self.pygit_repo
        commit1 = repo.get(commit)
        commit2 = self.commit_before_commit(commit)
        if commit2 is not None:
            commit2 = repo.get(commit2)

        diff = repo.diff(commit1, commit2)

        file_names = set()
        for patch in diff:
            delta = patch.delta
            file_names.add(delta.new_file.path)
            file_names.add(delta.old_file.path)
        return list(file_names)

    def list_commits(self, branch: Optional[str] = "") -> List[str]:
        repo = self.pygit_repo
        if branch:
            branch = repo.branches.get(branch)
            latest_commit_id = branch.target
        else:
            latest_commit_id = repo[repo.head.target].id

        return [commit for commit in repo.walk(latest_commit_id, pygit2.GIT_SORT_NONE)]

    def list_commit_shas(self, branch: Optional[str] = None) -> List[str]:
        branch = branch or self.default_branch
        return self.list_commits(branch)

    def list_n_commits(
        self,
        number: Optional[int] = 10,
        start_commit_sha: Optional[str] = None,
        branch: Optional[str] = None,
    ) -> List[str]:
        """List the specified number of top commits of the current branch
        Optionally skip a number of commits from the top"""
        if number <= 0:
            return []
        repo = self.pygit_repo
        if start_commit_sha is not None:
            start_commit_id = repo.get(start_commit_sha).id
        elif branch:
            branch = repo.branches.get(branch)
            start_commit_id = branch.target
        else:
            start_commit_id = repo[repo.head.target].id
        commits = itertools.islice(repo.walk(start_commit_id), number + 1)
        return [commit.hex for commit in commits if commit.hex != start_commit_sha]

    def list_modified_files(
        self, path: Optional[str] = None, with_status: Optional[bool] = False
    ) -> List[str]:
        # TODO
        # repo.diff("HEAD")
        # go over the returned diff object
        diff_command = "diff --name-status"
        if path is not None:
            diff_command = f"{diff_command} {path}"
        modified_files = self._git(diff_command).split("\n")
        files = []
        for modified_file in modified_files:
            # ignore warning lines
            if len(modified_file) and modified_file[0] in ["A", "M", "D"]:
                if with_status:
                    files.append(tuple(modified_file.split(maxsplit=1)))
                else:
                    files.append(modified_file.split(maxsplit=1)[1])
        return files

    def list_tags(self) -> List[str]:
        return self._git("tag -l").splitlines()

    def list_untracked_files(self, path: Optional[str] = None) -> List[str]:
        ls_command = "ls-files --others"
        if path is not None:
            ls_command = f"{ls_command} {path}"
        untracked_files = self._git(ls_command).split("\n")
        return [
            untracked_file for untracked_file in untracked_files if len(untracked_file)
        ]

    def list_worktrees(self) -> Dict[Path, Tuple[Path, str, str]]:
        """
        Returns a dictionary containing information about repository's worktrees:
        {
            "worktree1_path: (worktree1_path, worktree1_commit, worktree1_branch),
            "worktree2_path: (worktree2_path, worktree2_commit, worktree2_branch),
            ...
        }
        """
        worktrees_list = self._git("worktree list")
        worktrees = [w.split() for w in worktrees_list.splitlines() if w]
        return {
            Path(wt[0]): (Path(wt[0]), wt[1], wt[2].replace("[", "").replace("]", ""))
            for wt in worktrees
        }

    def merge_commit(
        self,
        commit: str,
        fast_forward_only: Optional[bool] = False,
        check_if_merge_completed: Optional[bool] = False,
    ) -> bool:
        fast_forward_only_flag = "--ff-only" if fast_forward_only else ""
        self._git("merge {} {}", commit, fast_forward_only_flag, log_error=True)
        if check_if_merge_completed:
            try:
                self._git("rev-parse -q --verify MERGE_HEAD")
                return False
            except GitError:
                pass
        return True

    def merge_branch(
        self, branch_name: str, allow_new_commit: Optional[bool] = False
    ) -> None:
        if allow_new_commit:
            repo = self.pygit_repo
            branch = repo.lookup_branch(branch_name)
            oid = branch.target

            for commit in repo.walk(oid):
                message = commit.message
                break

            repo.merge(oid)
            self.commit(message)
            repo.state_cleanup()
        else:
            self._git("merge {}", branch_name, log_error=True)

    def pull(self):
        """Pull current branch"""
        self._git("pull", log_error=True)

    def push(
        self,
        branch: Optional[str] = None,
        set_upstream: Optional[bool] = False,
        force: Optional[bool] = False,
    ) -> None:
        """Push all changes"""
        if branch is None:
            branch = self.get_current_branch()

        upstream_flag = "-u" if set_upstream else ""
        force_flag = "-f" if force else ""
        self._git(
            "push {} {} origin {}",
            upstream_flag,
            force_flag,
            branch,
            log_error=True,
            reraise_error=True,
            log_success_msg="Successfully pushed commit(s)",
        )

    def remove_remote(self, remote_name: str) -> None:
        try:
            self._git("remote remove {}", remote_name)
        except GitError as e:
            if "No such remote" not in str(e):
                raise

    def rename_branch(self, old_name: str, new_name: str) -> None:
        self._git("branch -m {} {}", old_name, new_name)

    def reset_num_of_commits(
        self, num_of_commits: int, hard: Optional[bool] = False
    ) -> None:
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} HEAD~{num_of_commits}")

    def reset_to_commit(self, commit: str, hard: Optional[bool] = False) -> None:
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} {commit}")

    def reset_to_head(self) -> None:
        self._git("reset --hard HEAD")

    def safely_get_json(self, commit: str, path: str) -> Optional[Dict]:
        try:
            return self.get_json(commit, path)
        except GitError:
            self._log_debug(f"{path} not available at revision {commit}")
        except json.decoder.JSONDecodeError:
            self._log_debug(f"{path} not a valid json at revision {commit}")
        return None

    def set_remote_url(self, new_url: str, remote: Optional[str] = "origin") -> None:
        self._git(f"remote set-url {remote} {new_url}")

    def set_upstream(self, branch_name: str) -> None:
        self._git("branch -u origin/{}", branch_name)

    def something_to_commit(self) -> bool:
        """Checks if there are any uncommitted changes"""
        uncommitted_changes = self._git("status --porcelain")
        return bool(uncommitted_changes)

    def synced_with_remote(
        self, branch: Optional[str] = None, url: Optional[str] = None
    ) -> bool:
        """Checks if local branch is synced with its remote branch"""
        # check if the latest local commit matches
        # the latest remote commit on the specified branch
        urls = (
            [url]
            if url is not None
            else self.urls
            if self.urls
            else [self.get_remote_url()]
        )
        for url in urls:
            branch = branch or self.default_branch
            tracking_branch = self.get_tracking_branch(branch, strip_remote=True)
            if not tracking_branch:
                return False
            try:
                local_commit = self.top_commit_of_branch(branch)
            except GitError as e:
                if "unknown revision or path not in the working tree" not in str(e):
                    raise e
                local_commit = None

        remote_commit = self.get_last_remote_commit(url, tracking_branch)

        return local_commit == remote_commit

    def top_commit_of_branch(self, branch_name: str) -> Optional[str]:
        repo = self.pygit_repo
        branch = repo.branches.get(branch_name)
        if branch is not None:
            return branch.target.hex
        # a reference like HEAD
        try:
            return repo.revparse_single(branch_name).id.hex
        except Exception:
            return None

    def _determine_default_branch(self) -> Optional[str]:
        """Determine the default branch of the repository"""
        # try to get the default branch from the local repository
        errors = []
        try:
            return self.get_default_branch()
        except GitError as e:
            errors.append(e)
            pass

        # if the local repository does not exist or doesn't have a default branch, try to get the default branch from remote
        if self.urls:
            for url in self.urls:
                try:
                    return self.get_default_branch(url)
                except GitError as e:
                    errors.append(e)
                    pass

        self._log_debug(
            f"Cannot determine default branch with git -C at {self.path}: {errors}"
        )
        return None

    def _validate_repo_name(self, name: str) -> str:
        """Ensure the repo name is not malicious"""
        match = _repo_name_re.match(name)
        if not match:
            taf_logger.error(f"Repository name {name} is not valid")
            raise InvalidRepositoryError(
                "Repository name must be in format namespace/repository "
                "and can only contain letters, numbers, underscores and "
                f'dashes, but got "{name}"'
            )
        return name

    def _validate_repo_path(
        self, library_dir: str | Path, name: str, path: str = None
    ) -> Path:
        """
        validate repo path
        (since this is coming from potentially untrusted data)
        """
        # using OS to avoid resolve adding drive letter name on Windows
        repo_dir = os.path.join(str(library_dir), name or "")
        repo_dir = os.path.normpath(repo_dir)
        if not repo_dir.startswith(str(Path(library_dir))):
            taf_logger.error(f"Repository path/name {library_dir}/{name} is not valid")
            raise InvalidRepositoryError(
                f"Repository path/name {library_dir}/{name} is not valid"
            )
        repo_dir = Path(repo_dir).expanduser().resolve()
        if path is not None and path != repo_dir:
            raise InvalidRepositoryError(
                "Both library dir and name pair and path specified and are not equal. Omit the path."
            )
        return repo_dir

    def _validate_url(self, url: str) -> None:
        """ensure valid URL"""
        for _url_re in [_http_fttp_url, _ssh_url]:
            match = _url_re.match(url)
            if match:
                return
        self._log_error(f"URL ({url}) is not valid")
        raise InvalidRepositoryError(
            f'Repository URL must be a valid URL, but got "{url}".'
        )

    def _validate_urls(self, urls: List[str]) -> List[str]:
        def _find_url(path, url):
            try:
                if (path / url).resolve().is_dir():
                    return str((path / url).resolve())
            except OSError:
                pass
            if os.path.isabs(url):
                return url
            return str((path).resolve())

        if urls is not None:
            if settings.update_from_filesystem is False:
                for url in urls:
                    self._validate_url(url)
            else:
                urls = [_find_url(self.path, url) for url in urls]
        return urls


_remote_branch_re = re.compile(r"^(refs/heads/)")

_local_branch_re = re.compile(r"^(origin/)")

_repo_name_re = re.compile(r"^\w[\w_-]*/\w[\w_-]*$")

_http_fttp_url = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    # domain...
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)

_ssh_url = re.compile(
    r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)?(/)?"
)



import os
import sys
import logging
import securesystemslib
from pathlib import Path

import tuf.log
import tuf.repository_tool
import tuf.exceptions
from loguru import logger as taf_logger

import taf.settings as settings

_CONSOLE_FORMAT_STRING = "\n{message}\n"
_FILE_FORMAT_STRING = "[{time}] [{level}] [{module}:{function}@{line}]\n{message}\n"

console_loggers = {}
file_loggers = {}


def disable_console_logging():
    taf_logger.remove(console_loggers["log"])
    disable_tuf_console_logging()


def disable_file_logging():
    taf_logger.remove(file_loggers["log"])
    disable_tuf_console_logging()


def disable_tuf_console_logging():
    try:
        tuf.log.set_console_log_level(logging.CRITICAL)
    except securesystemslib.exceptions.Error:
        pass


def disable_tuf_file_logging():
    if tuf.log.file_handler is not None:
        tuf.log.disable_file_logging()
    else:
        logging.getLogger("tuf").setLevel(logging.CRITICAL)
    logging.getLogger("securesystemslib_keys").setLevel(logging.CRITICAL)
    logging.getLogger("securesystemslib_util").setLevel(logging.CRITICAL)


def _get_log_location():
    location = settings.LOGS_LOCATION or os.environ.get("TAF_LOG")
    if location is None:
        location = Path.home() / ".taf"
        location.mkdir(exist_ok=True)
    else:
        location = Path(location)
    return location


taf_logger.remove()

if settings.ENABLE_CONSOLE_LOGGING:
    console_loggers["log"] = taf_logger.add(
        sys.stdout, format=_CONSOLE_FORMAT_STRING, level=settings.CONSOLE_LOGGING_LEVEL
    )
    tuf.log.set_console_log_level(settings.CONSOLE_LOGGING_LEVEL)
else:
    # if console logging is disable, remove tuf console logger
    disable_tuf_console_logging()


if settings.ENABLE_FILE_LOGGING:
    logs_location = _get_log_location()
    log_path = str(logs_location / settings.LOG_FILENAME)
    file_loggers["log"] = taf_logger.add(
        log_path, format=_FILE_FORMAT_STRING, level=settings.FILE_LOGGING_LEVEL
    )

    if settings.SEPARATE_ERRORS:
        error_log_path = str(logs_location / settings.ERROR_LOG_FILENAME)
        file_loggers["error"] = taf_logger.add(
            error_log_path,
            format=_FILE_FORMAT_STRING,
            level=settings.ERROR_LOGGING_LEVEL,
        )
    try:
        tuf.log.set_filehandler_log_level(settings.FILE_LOGGING_LEVEL)
    except tuf.exceptions.Error:
        pass
else:
    # if file logging is disabled, also disable tuf file logging
    disable_tuf_file_logging()



# Default scheme for all RSA keys. It can be changed in keys.json while
# generating repository
import datetime


DEFAULT_RSA_SIGNATURE_SCHEME = "rsa-pkcs1v15-sha256"

CAPSTONE = "capstone"

DEFAULT_ROLE_SETUP_PARAMS = {
    "number": 1,
    "threshold": 1,
    "yubikey": False,
    "scheme": DEFAULT_RSA_SIGNATURE_SCHEME,
    "length": 3072,
    "passwords": None,
    "terminating": True,
}

# Yubikey x509 certificate expiration interval
EXPIRATION_INTERVAL = 36500

YUBIKEY_EXPIRATION_DATE = datetime.datetime.now() + datetime.timedelta(
    days=EXPIRATION_INTERVAL
)



import click
from taf.api.roles import add_role, add_roles, list_keys_of_role, remove_role, add_signing_key as add_roles_signing_key
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.tools.cli import catch_cli_exception


def attach_to_group(group):

    @group.group()
    def roles():
        pass

    @roles.command()
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--parent-role", default="targets", help="Parent targets role of this role. Defaults to targets")
    @click.option("--delegated-path", multiple=True, help="Paths associated with the delegated role")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--keys-number", default=1, help="Number of signing keys. Defaults to 1")
    @click.option("--threshold", default=1, help="An integer number of keys of that "
                  "role whose signatures are required in order to consider a file as being properly signed by that role")
    @click.option("--yubikey", is_flag=True, default=None, help="A flag determining if the new role should be signed using a Yubikey")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be "
                  "committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not "
                  "located inside the keystore directory")
    def add(role, path, parent_role, delegated_path, keystore, keys_number, threshold, yubikey, scheme, no_commit, prompt_for_keys):
        """Add a new delegated target role, specifying which paths are delegated to the new role.
        Its parent role, number of signing keys and signatures threshold can also be defined.
        Update and sign all metadata files and commit.
        """
        if not path:
            print("Specify at least one path")
            return

        add_role(
            path=path,
            role=role,
            parent_role=parent_role,
            paths=delegated_path,
            keys_number=keys_number,
            threshold=threshold,
            yubikey=yubikey,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
        )

    @roles.command()
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.argument("keys-description")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme "
                  "used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be "
                  "committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not "
                  "located inside the keystore directory")
    def add_multiple(path, keystore, keys_description, scheme, no_commit, prompt_for_keys):
        """Add one or more target roles. Information about the roles
        can be provided through a dictionary - either specified directly or contained
        by a .json file whose path is specified when calling this command. This allows
        definition of:
            - total number of keys per role
            - threshold of signatures per role
            - should keys of a role be on Yubikeys or should keystore files be used
            - scheme (the default scheme is rsa-pkcs1v15-sha256)
            - keystore path, if not specified via keystore option

        \b
        For example:
        {
            "roles": {
                "root": {
                    "number": 3,
                    "length": 2048,
                    "passwords": ["password1", "password2", "password3"],
                    "threshold": 2,
                    "yubikey": true
                },
                "targets": {
                    "length": 2048
                },
                "snapshot": {},
                "timestamp": {}
            },
            "keystore": "keystore_path"
        }
        """
        add_roles(
            path=path,
            keystore=keystore,
            roles_key_infos=keys_description,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
            commit=not no_commit,
        )

    @roles.command()
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme "
                  "used for signing")
    @click.option("--remove-targets/--no-remove-targets", default=True, help="Should targets delegated to this "
                  "role also be removed. If not removed, they are signed by the parent role")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be "
                  "committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not "
                  "located inside the keystore directory",)
    def remove(role, path, keystore, scheme, remove_targets, no_commit, prompt_for_keys):
        """Remove a delegated target role, and, optionally, its targets (depending on the remove-targets parameter).
        If targets should also be deleted, target files are remove and their corresponding entires are removed
        from repositoires.json. If targets should not get removed, the target files are signed using the
        removed role's parent role
        """
        remove_role(
            path=path,
            role=role,
            keystore=keystore,
            scheme=scheme,
            remove_targets=remove_targets,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
        )

    @roles.command()
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles to whose list of signing keys "
                  "the new key should be added")
    @click.option("--pub-key-path", default=None, help="Path to the public key corresponding to "
                  "the private key which should be registered as the role's signing key")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--keys-description", help="A dictionary containing information about the "
                  "keys or a path to a json file which stores the needed information")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme "
                  "used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be "
                  "committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not "
                  "located inside the keystore directory")
    def add_signing_key(path, role, pub_key_path, keystore, keys_description, scheme, no_commit, prompt_for_keys):
        """
        Add a new signing key. This will make it possible to a sign metadata files
        corresponding to the specified roles with another key. Although private keys are
        used for signing, key identifiers are calculated based on the public keys. This
        means that it's necessary to enter the public key in order to register
        a new signing key. Public key can be loaded from a file, in which case it is
        necessary to specify its path as the pub_key parameter's value. If this option
        is not used when calling this command, the key can be directly entered later.
        """
        if not len(role):
            print("Specify at least one role")
            return
        add_roles_signing_key(
            path=path,
            roles=role,
            pub_key_path=pub_key_path,
            keystore=keystore,
            roles_key_infos=keys_description,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys
        )

    @roles.command()
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    def list_keys(role, path):
        """
        List all keys of the specified role. If certs directory exists and contains certificates exported from YubiKeys,
        include additional information read from these certificates, like name or organization.
        """
        list_keys_of_role(
            path=path,
            role=role,
        )



from logging import DEBUG, ERROR, INFO
import os
from typing import Dict, List, Optional
import click
from collections import defaultdict
from functools import partial
import json
from pathlib import Path
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.api.utils import check_if_clean, commit_and_push
from taf.exceptions import TAFError
from taf.models.converter import from_dict
from taf.models.types import RolesIterator
from taf.models.types import DelegatedRole, Role, TargetsRole
from taf.repositoriesdb import REPOSITORIES_JSON_PATH
from taf.yubikey import get_key_serial_by_id
from tuf.repository_tool import TARGETS_DIRECTORY_NAME, Metadata
import tuf.roledb
import taf.repositoriesdb as repositoriesdb
from taf.keys import (
    get_key_name,
    get_metadata_key_info,
    load_signing_keys,
    load_sorted_keys_of_new_roles,
)
from taf.api.metadata import update_snapshot_and_timestamp, update_target_metadata
from taf.auth_repo import AuthenticationRepository
from taf.constants import (
    DEFAULT_ROLE_SETUP_PARAMS,
    YUBIKEY_EXPIRATION_DATE,
    DEFAULT_RSA_SIGNATURE_SCHEME,
)
from taf.keystore import default_keystore_path, new_public_key_cmd_prompt
from taf.repository_tool import (
    Repository,
    is_delegated_role,
    yubikey_signature_provider,
)
from taf.utils import get_key_size, read_input_dict
from taf.log import taf_logger
from taf.models.types import RolesKeysData


MAIN_ROLES = ["root", "snapshot", "timestamp", "targets"]


@log_on_start(DEBUG, "Adding a new role {role:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding a new role", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding a new role {role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_role(
    path: str,
    role: str,
    parent_role: str,
    paths: list,
    keys_number: int,
    threshold: int,
    yubikey: bool,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    auth_repo: Optional[AuthenticationRepository] = None,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Add a new delegated target role and update and sign metadata files.
    Automatically commit the changes if commit is set to True.

    Arguments:
        path: Path to the authentication repository.
        role: Name of the role which is to be added.
        parent_role: Name of the target role that is the new role's parent. Can be targets or another delegated role.
        paths: A list of target paths that are delegated to the new role.
        keys_number: Total number of signing keys
        threshold: Signature's threshold.
        yubikey: Specifies if the metadata file should be signed using a YubiKey.
        keystore: Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.

    Side Effects:
        Initializes a new delegated targets role, signs metadata files, write changes to the disk and optionally commits.

    Returns:
        None
    """
    if auth_repo is None:
        auth_repo = AuthenticationRepository(path=path)
    path = Path(path)
    existing_roles = auth_repo.get_all_targets_roles()
    existing_roles.extend(MAIN_ROLES)
    if role in existing_roles:
        taf_logger.info("All roles already set up")
        return

    if parent_role == "targets":
        parent_role = TargetsRole()
    else:
        parent_role = DelegatedRole(name=parent_role)

    new_role = DelegatedRole(
        parent=parent_role,
        name=role,
        paths=paths,
        number=keys_number,
        threshold=threshold,
        yubikey=yubikey,
    )

    signing_keys, verification_keys = load_sorted_keys_of_new_roles(
        auth_repo=auth_repo,
        roles=new_role,
        yubikeys_data=None,
        keystore=keystore,
    )
    create_delegations(
        new_role, auth_repo, verification_keys, signing_keys, existing_roles
    )
    _update_role(
        auth_repo,
        parent_role.name,
        keystore,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
    )
    if commit:
        update_snapshot_and_timestamp(
            auth_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
        )
        commit_and_push(auth_repo)
    else:
        taf_logger.info("\nPlease commit manually\n")


@log_on_start(DEBUG, "Adding new paths to role {delegated_role:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding new paths to role", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding new paths to role {delegated_role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def add_role_paths(
    paths: List[str],
    delegated_role: str,
    keystore: str,
    commit: Optional[bool] = True,
    auth_repo: Optional[AuthenticationRepository] = None,
    auth_path: Optional[str] = None,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Adds additional delegated target paths to the specified role. That means that
    the role will be responsible for sining target files at those location going forward.

    Arguments:
        paths: A list of additional target paths that should be delegated to the role.
        delegated_role: Name of the targets role to which the new paths should be delegated.
        auth_path: Path to the authentication repository.
        keystore: Location of the keystore files.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
    Side Effects:
        Updates the specified target role's parent and other metadata files (snapshot and timestamp),
        signs them, writes changes to disk and optionally commits everything.

    Returns:
        None
    """
    if auth_repo is None:
        auth_repo = AuthenticationRepository(path=auth_path)
    parent_role = auth_repo.find_delegated_roles_parent(delegated_role)
    parent_role_obj = _role_obj(parent_role, auth_repo)
    parent_role_obj.add_paths(paths, delegated_role)
    _update_role(auth_repo, parent_role, keystore, prompt_for_keys=prompt_for_keys)
    if commit:
        update_snapshot_and_timestamp(
            auth_repo, keystore, prompt_for_keys=prompt_for_keys
        )
        commit_and_push(auth_repo)
    else:
        taf_logger.info("\nPlease commit manually\n")


@log_on_start(DEBUG, "Adding new roles", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding new roles", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding new roles: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_roles(
    path: str,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    commit: Optional[bool] = True,
) -> None:
    """
    Add new target roles and sign all metadata files given information stored in roles_key_infos
    dictionary or .json file

    Arguments:
        path: Path to the authentication repository.
        keystore (optional): Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.

    Side Effects:
        Updates metadata files (parent of new roles, snapshot and timestamp) and creates new targets metadata files.
        Writes changes to disk.

    Returns:
        None
    """
    auth_repo = AuthenticationRepository(path=path)
    path = Path(path)

    roles_key_infos, keystore = _initialize_roles_and_keystore(
        roles_key_infos, keystore
    )

    roles_keys_data = from_dict(roles_key_infos, RolesKeysData)

    new_roles = []
    existing_roles = auth_repo.get_all_targets_roles()
    main_roles = ["root", "snapshot", "timestamp"]
    existing_roles.extend(main_roles)

    new_roles = [
        role
        for role in RolesIterator(roles_keys_data.roles.targets, skip_top_role=True)
        if role.name not in existing_roles
    ]

    parent_roles_names = {role.parent.name for role in new_roles}

    if not len(new_roles):
        taf_logger.info("All roles already set up")
        return

    repository = auth_repo._repository
    signing_keys, verification_keys = load_sorted_keys_of_new_roles(
        auth_repo=auth_repo,
        roles=roles_keys_data.roles,
        keystore=keystore,
        yubikeys_data=roles_keys_data.yubikeys,
        existing_roles=existing_roles,
    )

    create_delegations(
        roles_keys_data.roles.targets,
        repository,
        verification_keys,
        signing_keys,
        existing_roles,
    )
    for parent_role_name in parent_roles_names:
        _update_role(
            auth_repo,
            parent_role_name,
            keystore,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
        )
    update_snapshot_and_timestamp(
        auth_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    )
    if commit:
        commit_and_push(auth_repo)
    else:
        taf_logger.info("\nPlease commit manually\n")


@log_on_start(DEBUG, "Adding new signing key to roles", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding new signing key to roles", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding new signing key to roles: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_signing_key(
    path: str,
    roles: List[str],
    pub_key_path: Optional[str] = None,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Add a new signing key to the listed roles. Update root metadata if one or more roles is one of the main TUF roles,
    parent target role if one of the roles is a delegated target role and timestamp and snapshot in any case.

    Arguments:
        path: Path to the authentication repository.
        roles: A list of roles whose signing keys need to be extended.
        pub_key_path (optional): path to the file containing the public component of the new key. If not provided,
            it will be necessary to ender the key when prompted.
        keystore (optional): Location of the keystore files.
        roles_key_infos (optional): Path to a json file which contains information about repository's roles and keys.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.

    Side Effects:
        Updates metadata files (parents of the affected roles, snapshot and timestamp).
        Writes changes to disk.

    Returns:
        None
    """
    auth_repo = AuthenticationRepository(path=path)
    roles_key_infos, keystore = _initialize_roles_and_keystore(
        roles_key_infos, keystore, enter_info=False
    )

    pub_key_pem = None
    if pub_key_path is not None:
        pub_key_path = Path(pub_key_path)
        if pub_key_path.is_file():
            pub_key_pem = Path(pub_key_path).read_text()

    if pub_key_pem is None:
        pub_key_pem = new_public_key_cmd_prompt(scheme)["keyval"]["public"]

    parent_roles = set()
    for role in roles:
        if auth_repo.is_valid_metadata_key(role, pub_key_pem):
            taf_logger.info(f"Key already registered as signing key of role {role}")
            continue

        auth_repo.add_metadata_key(role, pub_key_pem, scheme)

        if is_delegated_role(role):
            parent_role = auth_repo.find_delegated_roles_parent(role)
        else:
            parent_role = "root"

        parent_roles.add(parent_role)

    if not len(parent_roles):
        return

    auth_repo.unmark_dirty_roles(list(set(roles) - parent_roles))
    for parent_role in parent_roles:
        _update_role(
            auth_repo,
            parent_role,
            keystore,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
        )

    update_snapshot_and_timestamp(
        auth_repo, keystore=keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    )

    if commit:
        commit_and_push(auth_repo)
    else:
        taf_logger.info("\nPlease commit manually\n")


def _enter_roles_infos(keystore: str, roles_key_infos: str) -> Dict:
    """
    Ask the user to enter information taf roles and keys, including the location
    of keystore directory if not entered through an input parameter

    Arguments:
        keystore: Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.

    Side Effects:
        None

    Returns:
        A dictionary containing entered information about taf roles and keys (total number of keys per role,
        parent roles of roles, threshold of signatures per role, indicator if metadata should be signed using
        a yubikey for each role, key length and signing scheme for each role)
    """
    mandatory_roles = ["root", "targets", "snapshot", "timestamp"]
    role_key_infos = defaultdict(dict)
    infos_json = {}

    for role in mandatory_roles:
        role_key_infos[role] = _enter_role_info(role, role == "targets", keystore)
    infos_json["roles"] = role_key_infos

    while keystore is None:
        keystore = input(
            "Enter keystore location if keys should be loaded from or generated to keystore files (leave empty otherwise): "
        )
        if len(keystore):
            if not Path(keystore).is_dir():
                taf_logger.info(f"{keystore} does not exist")
                keystore = None

    if keystore:
        infos_json["keystore"] = keystore

    def _print_roles_key_infos(infos_json_str):
        print("------------------")
        print(
            "Configuration json - save it in order to make creation of repositories quicker"
        )
        print(json.dumps(infos_json, indent=4))
        print("------------------")

    infos_json_str = json.dumps(infos_json, indent=4)
    if isinstance(roles_key_infos, str):
        try:
            path = Path(roles_key_infos)
            path.parent.mkdir(parents=True, exist_ok=True)
            Path(roles_key_infos).write_text(infos_json_str)
            taf_logger.info(
                f"Configuration json written to {Path(roles_key_infos).absolute()}"
            )
        except Exception as e:
            taf_logger.error(e)
            _print_roles_key_infos(infos_json_str)
    else:
        print(infos_json_str)
    return infos_json


def _enter_role_info(role: str, is_targets_role: bool, keystore: str) -> Dict:
    def _read_val(input_type, name, param=None, required=False):
        default_value_msg = ""
        if param is not None:
            default = DEFAULT_ROLE_SETUP_PARAMS.get(param)
            if default is not None:
                default_value_msg = f"(default {DEFAULT_ROLE_SETUP_PARAMS[param]}) "

        while True:
            try:
                val = input(f"Enter {name} and press ENTER {default_value_msg}")
                if not val:
                    if not required:
                        return DEFAULT_ROLE_SETUP_PARAMS.get(param)
                    else:
                        continue
                return input_type(val)
            except ValueError:
                pass

    role_info = {}
    keys_num = _read_val(int, f"number of {role} keys", "number")
    if keys_num is not None:
        role_info["number"] = keys_num

    role_info["yubikey"] = click.confirm(f"Store {role} keys on Yubikeys?")
    if role_info["yubikey"]:
        # in case of yubikeys, length and shceme have to have specific values
        role_info["length"] = 2048
        role_info["scheme"] = DEFAULT_RSA_SIGNATURE_SCHEME
    else:
        # if keystore is specified and contain keys corresponding to this role
        # get key size based on the public key
        keystore_length = _get_roles_key_size(role, keystore, keys_num)
        if keystore_length == 0:
            key_length = _read_val(int, f"{role} key length", "length")
            if key_length is not None:
                role_info["length"] = key_length
        else:
            role_info["length"] = keystore_length
        scheme = _read_val(str, f"{role} signature scheme", "scheme")
        if scheme is not None:
            role_info["scheme"] = scheme

    threshold = _read_val(int, f"{role} signature threshold", "threshold")
    if threshold is not None:
        role_info["threshold"] = threshold

    if is_targets_role:
        delegated_roles = defaultdict(dict)
        while click.confirm(
            f"Add {'another' if len(delegated_roles) else 'a'} delegated targets role of role {role}?"
        ):
            role_name = _read_val(str, "role name", True)
            delegated_paths = []
            while not len(delegated_paths) or click.confirm("Enter another path?"):
                delegated_paths.append(
                    _read_val(
                        str, f"path or glob pattern delegated to {role_name}", True
                    )
                )
            delegated_roles[role_name]["paths"] = delegated_paths
            is_terminating = click.confirm(f"Is {role_name} terminating?")
            delegated_roles[role_name]["terminating"] = is_terminating
            delegated_roles[role_name].update(
                _enter_role_info(role_name, True, keystore)
            )
        role_info["delegations"] = delegated_roles

    return role_info


def _initialize_roles_and_keystore(
    roles_key_infos: str, keystore: str, enter_info: Optional[bool] = True
):
    """
    Read information about roles and keys from a json file or ask the user to enter
    this information if not specified through a json file enter_info is True

    Arguments:
        roles_key_infos: A dictionary containing information about the roles:
            - total number of keys per role
            - their parent roles
            - threshold of signatures per role
            - should keys of a role be on Yubikeys or should a keystore files be used
            - scheme (the default scheme is rsa-pkcs1v15-sha256)
            - keystore path, if not specified via keystore option
        keystore: Location of the keystore files.
        enter_info (optional): Indicates if the user should be asked to enter information about the.
        roles and keys if not specified. Set to True by default.

    Side Effects:
        None

    Returns:
        A dictionary containing entered information about taf roles and keys (total number of keys per role,
        parent roles of roles, threshold of signatures per role, indicator if metadata should be signed using
        a yubikey for each role, key length and signing scheme for each role) and keystore file path
    """
    roles_key_infos_dict = read_input_dict(roles_key_infos)
    if keystore is None:
        # if keystore path is specified in roles_key_infos and is a relative path
        # it should be relative to the location of the file
        # roles_key_infos can either be path to a json file, or a dictionary (or not provided)
        keystore = roles_key_infos_dict.get("keystore") or default_keystore_path()
        if roles_key_infos is not None and type(roles_key_infos) == str:
            roles_key_infos_path = Path(roles_key_infos)
            if roles_key_infos_path.is_file() and "keystore" in roles_key_infos_dict:
                keystore_path = (
                    Path(roles_key_infos_dict["keystore"]).expanduser().resolve()
                )
                if not keystore_path.is_absolute():
                    keystore_path = (
                        roles_key_infos_path.parent / keystore_path
                    ).resolve()
                    keystore = str(keystore_path)

    if enter_info and not len(roles_key_infos_dict):
        # ask the user to enter roles, number of keys etc.
        roles_key_infos_dict = _enter_roles_infos(keystore, roles_key_infos)

    return roles_key_infos_dict, keystore


@log_on_start(INFO, "Creating delegations", logger=taf_logger)
@log_on_end(DEBUG, "Finished creating delegations", logger=taf_logger)
def create_delegations(
    role: str,
    repository: AuthenticationRepository,
    verification_keys: Dict,
    signing_keys: Dict,
    existing_roles: Optional[bool] = None,
) -> None:
    """
    Initialize new delegated target roles, update authentication repository object

    Arguments:
        role: Targets main role or a delegated role
        repository: Authentication repository.
        verification_keys: A dictionary containing mappings of role names to their verification (public) keys.
        signing_keys: A dictionary containing mappings of role names to their signing (private) keys.
        existing_roles: A list of already initialized roles.

    Side Effects:
        Updates authentication repository object

    Returns:
        None
    """
    if existing_roles is None:
        existing_roles = []
    skip_top_role = isinstance(role, TargetsRole)
    for delegated_role in RolesIterator(role, skip_top_role=skip_top_role):
        parent_role_obj = _role_obj(delegated_role.parent.name, repository)
        if delegated_role.name in existing_roles:
            taf_logger.info(f"Role {delegated_role.name} already set up.")
            continue
        paths = delegated_role.paths
        roles_verification_keys = verification_keys[delegated_role.name]
        # if yubikeys are used for signing, signing keys are not loaded
        roles_signing_keys = signing_keys.get(delegated_role.name)
        parent_role_obj.delegate(
            delegated_role.name,
            roles_verification_keys,
            paths,
            threshold=delegated_role.threshold,
            terminating=delegated_role.terminating,
        )
        setup_role(
            delegated_role,
            repository,
            roles_verification_keys,
            roles_signing_keys,
            parent=parent_role_obj,
        )


def _get_roles_key_size(role: str, keystore: str, keys_num: int) -> int:
    pub_key_name = f"{get_key_name(role, 1, keys_num)}.pub"
    key_path = str(Path(keystore, pub_key_name))
    return get_key_size(key_path)


def _role_obj(role: str, repository: Repository, parent: str = None) -> Metadata:
    """
    Return role TUF object based on its name
    """
    if isinstance(repository, Repository):
        repository = repository._repository
    if role == "targets":
        return repository.targets
    elif role == "snapshot":
        return repository.snapshot
    elif role == "timestamp":
        return repository.timestamp
    elif role == "root":
        return repository.root
    else:
        # return delegated role
        if parent is None:
            return repository.targets(role)
        return parent(role)


@log_on_error(
    ERROR,
    "Could not list keys of {role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def list_keys_of_role(
    path: str,
    role: str,
) -> None:
    """
     Print information about signing keys of role. If a certificate whose name matches
     a key's id exists, include information contained by that certificate (like name and valid to/from dates)

     Arguments:
         path: Path to the authentication repository.
         role: Name of the role which is to be removed.

    Side Effects:
         None

     Returns:
         None
    """
    auth_repo = AuthenticationRepository(path=path)
    key_ids = auth_repo.get_role_keys(role=role)
    for key_id in key_ids:
        taf_logger.info(f"{get_metadata_key_info(auth_repo.certs_dir, key_id)}")


@log_on_start(DEBUG, "Removing role {role:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished removing the role", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while removing role {role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def remove_role(
    path: str,
    role: str,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    remove_targets: Optional[bool] = False,
    auth_repo: Optional[AuthenticationRepository] = None,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Remove a delegated target role and update and sign metadata files.
    Automatically commit the changes if commit is set to True.
    It is not possible to remove any of the main TUF roles

    Arguments:
        path: Path to the authentication repository.
        role: Name of the role which is to be removed.
        keystore: Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        remove_targets (optional): Indicates if target files should be removed to, or signed by the parent role.
            Set to False by default.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.

    Side Effects:
        Updates metadata files, optionally deletes target files, writes changes to disk and optionally commits.

    Returns:
        None
    """
    if role in MAIN_ROLES:
        taf_logger.info(
            f"Cannot remove role {role}. It is one of the roles required by the TUF specification"
        )
        return

    if auth_repo is None:
        auth_repo = AuthenticationRepository(path=path)

    parent_role = auth_repo.find_delegated_roles_parent(role)
    if parent_role is None:
        taf_logger.info("Role is not among delegated roles")
        return

    roleinfo = tuf.roledb.get_roleinfo(parent_role, auth_repo.name)
    added_targets_data = {}
    removed_targets = []
    for delegations_data in roleinfo["delegations"]["roles"]:
        if delegations_data["name"] == role:
            paths = delegations_data["paths"]
            for path in paths:
                target_file_path = Path(path, TARGETS_DIRECTORY_NAME, path)
                if target_file_path.is_file():
                    if remove_targets:
                        os.unlink(str(target_file_path))
                        removed_targets.append(path)
                    else:
                        added_targets_data[path] = {}
            break

    parent_role_obj = _role_obj(parent_role, auth_repo)
    parent_role_obj.revoke(role)

    _update_role(
        auth_repo, role=parent_role, keystore=keystore, prompt_for_keys=prompt_for_keys
    )
    if len(added_targets_data):
        removed_targets_data = {}
        update_target_metadata(
            auth_repo,
            added_targets_data,
            removed_targets_data,
            keystore,
            write=False,
            scheme=DEFAULT_RSA_SIGNATURE_SCHEME,
            update_target_metadata=update_target_metadata,
            prompt_for_keys=prompt_for_keys,
        )

    # if targets should be deleted, also removed them from repositories.json
    if len(removed_targets):
        repositories_json = repositoriesdb.load_repositories_json(auth_repo)
        repositories = repositories_json["repositories"]
        for removed_target in removed_targets:
            if removed_target in repositories:
                repositories.pop(removed_target)

            # update content of repositories.json before updating targets metadata
            Path(auth_repo.path, REPOSITORIES_JSON_PATH).write_text(
                json.dumps(repositories_json, indent=4)
            )

    update_snapshot_and_timestamp(
        auth_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    )
    if commit:
        commit_and_push(auth_repo)
    else:
        taf_logger.info("Please commit manually")


@log_on_start(DEBUG, "Removing delegated paths", logger=taf_logger)
@log_on_end(DEBUG, "Finished removing delegated paths", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while removing roles: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def remove_paths(
    path: str,
    paths: List[str],
    keystore: str,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
) -> bool:
    """
    Remove delegated paths. Update parent roles of the roles associated with the removed paths,
    as well as snapshot and timestamp. Optionally commit the changes.

    Arguments:
        path:  Path to the authentication repository.
        paths: Paths to be removed.
        keystore: Location of the keystore files.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.

    Side Effects:
        Updates metadata files, writes changes to disk and optionally commits them.

    Returns:
        True if the delegation existed, False otherwise
    """
    auth_repo = AuthenticationRepository(path=path)
    delegation_existed = False
    for path_to_remove in paths:
        delegated_role = auth_repo.get_role_from_target_paths([path_to_remove])
        if delegated_role != "targets":
            parent_role = auth_repo.find_delegated_roles_parent(delegated_role)
            # parent_role_obj = _role_obj(parent_role, auth_repo)
            current_delegation_existed = _remove_path_from_role_info(
                path_to_remove, parent_role, delegated_role, auth_repo
            )
            delegation_existed = delegation_existed or current_delegation_existed
            if current_delegation_existed:
                _update_role(
                    auth_repo, parent_role, keystore, prompt_for_keys=prompt_for_keys
                )
    if delegation_existed and commit:
        update_snapshot_and_timestamp(
            auth_repo, keystore, prompt_for_keys=prompt_for_keys
        )
        commit_and_push(auth_repo)
    elif delegation_existed:
        taf_logger.info("\nPlease commit manually\n")
    return delegation_existed


def _remove_path_from_role_info(
    path_to_remove: str,
    parent_role: str,
    delegated_role: str,
    auth_repo: AuthenticationRepository,
) -> bool:
    """
    Remove path from delegated paths if directly listed.

    E.g. if delegated paths are

    "paths": [
      "namespace/repo1",
      "namespace/repo2"
    ]

    Arguments:
        path_to_remove: path to be removed from delegated paths
        parent_role: Parent role's name
        delegated_role: Delegated role's name
        auth_repo: Authentication repository

    and namespace/repo1 is being removed

    Returns:
        True if path was directly specified as a delegated path, False otherwise
    """

    auth_repo.reload_tuf_repository()
    delegation_exists = False
    roleinfo = tuf.roledb.get_roleinfo(parent_role, auth_repo.name)
    for delegations_data in roleinfo["delegations"]["roles"]:
        if delegations_data["name"] == delegated_role:
            delegations_paths = delegations_data["paths"]
            if path_to_remove in delegations_paths:
                delegations_paths.remove(path_to_remove)
                delegation_exists = True
            else:
                taf_logger.info(f"{path_to_remove} not in delegated paths")
            break
    if delegation_exists:
        tuf.roledb.update_roleinfo(
            parent_role, roleinfo, repository_name=auth_repo.name
        )
    return delegation_exists


@log_on_start(INFO, "Setting up role {role.name:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished setting up role {role.name:s}", logger=taf_logger)
def setup_role(
    role: Role,
    repository: Repository,
    verification_keys: Dict,
    signing_keys: Optional[Dict] = None,
    parent: Optional[str] = None,
) -> None:
    """
    Initialize a new role, add signing and verification keys.
    """
    role_obj = _role_obj(role.name, repository, parent)
    role_obj.threshold = role.threshold
    if not role.is_yubikey:
        for public_key, private_key in zip(verification_keys, signing_keys):
            role_obj.add_verification_key(public_key)
            role_obj.load_signing_key(private_key)
    else:
        for key_name, key in zip(role.yubikeys, verification_keys):
            role_obj.add_verification_key(key, expires=YUBIKEY_EXPIRATION_DATE)
            # check if yubikey loaded
            if get_key_serial_by_id(key_name):
                role_obj.add_external_signature_provider(
                    key, partial(yubikey_signature_provider, key_name, key["keyid"])
                )


def _update_role(
    auth_repo: AuthenticationRepository,
    role: str,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Update the specified role's metadata's expiration date, load the signing keys
    from either a keystore file or yubikey and sign the file without updating
    snapshot and timestamp and writing changes to disk
    """
    keystore_keys, yubikeys = load_signing_keys(
        auth_repo, role, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    )
    if len(keystore_keys):
        auth_repo.update_role_keystores(role, keystore_keys, write=False)
    if len(yubikeys):
        auth_repo.update_role_yubikeys(role, yubikeys, write=False)



import click
from functools import partial, wraps
from logging import ERROR
from logdecorator import log_on_error

from taf.exceptions import TAFError
from taf.log import taf_logger


def catch_cli_exception(func=None, *, handle, print_error=False):
    if not func:
        return partial(catch_cli_exception, handle=handle)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except handle as e:
            if print_error:
                click.echo(e)

    return wrapper


@log_on_error(
    ERROR,
    "{e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def process_custom_command_line_args(ctx):
    """
    Certain cli commands need to make it possible to enter custom arguments.
    E.g. when adding a new target repository, users can specify using the cli
    key-value pairs that will be written to its custom section of repositories.json.
    This function checks if that input is valid. It is easy to accidentally omit a space,
    e.g. --arg1 val1--arg2 val2. So show a nice error message in such a case. Also convert
    boolean values to actual booleans.

    Arguments:
        ctx: click context

    Returns:
       A dictionary containing parsed custom arguments
    """

    def _convert_value(value):
        if value.lower() == "true":
            return True
        if value.lower() == "false":
            return False
        return value

    if len(ctx.args) % 2 == 1:
        raise TAFError(
            "Custom parameters invalid. Check if there are spaces around each parameter/value"
        )

    custom = (
        {
            ctx.args[i][2:]: _convert_value(ctx.args[i + 1])
            for i in range(0, len(ctx.args), 2)
        }
        if len(ctx.args)
        else {}
    )
    return custom
